! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module:dss_bitmasks_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains auxiliary routines to do with the data subspace (DSS).
! 
! Comments: TODO:
! 
! $Id: dss_bitmasks_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module dss_bitmasks_aux

  use types
  use errorHandling
  use dss

  use constants		! for lnTwo.
  use ssc_misc		! for calcMostSignifSetBit, calcLeastSignifSetBit.
  use dss_ranges_aux	! for rangesAreWellFormed.

  implicit none 

  private 

  public getBitMasks
  public initializeBitMask
  public allBitsEquiv
  public convertRangeFilterToBitMask
  public bitMaskFilterIsBad
  public andBitMasks

  interface getBitMasks
    module procedure getBitMasksDssComp, getBitMasksDssFilt
  end interface

  interface andBitMasks
    module procedure andBitMasksScalVec,    andBitMasksScalScalVec
    module procedure andBitMasksScalVecVec, andBitMasksVecScalVec
    module procedure andBitMasksVecVec,     andBitMasksVecVecVec
    module procedure andBitMasksScalScal,   andBitMasksScalScalScal
  end interface

  type, public :: BitMaskT
    integer(int32) :: onBits, offBits
  end type BitMaskT

  integer, parameter :: &	! See dsslib/src/dsscore/dssinter.f90 for the definition of 80 and 255
       dssFiltCharLength = 80
    !, dssFilterBlockCharLength = 255

contains 

  !---------------------------------------------------------------------
  subroutine getBitMasksDssComp(dssComp, axisName, totalBitMasks)
  !---------------------------------------------------------------------
  ! The purpose of this subroutine is to return a vector of int32 bit masks representing the sum of all filters on the axisName column of this dss component. There are two parts to this: first, cycling through all the filters and converting each filter on FLAG into an int32 bit mask; secondly, ANDing these bit masks together.
  !
  ! NOTE! The calling routine should deallocate totalBitMasks after use.

    implicit none

    type(DScompT),  intent(in) :: dssComp
    character(*),   intent(in) :: axisName
    type(BitMaskT), pointer    :: totalBitMasks(:)	! intent(out)

    integer :: numFilts, filtNum
    type(DSfilterT) :: dssFilt
    type(BitMaskT), pointer :: filtBitMask(:) => null()
    character(dssFiltCharLength) :: tmpDssFiltName

    !-------------------------------------------------------------------
    ! Start

!print *, 'aaa getBitMasksDssComp'
!print *, 'axisName=', axisName
    allocate(totalBitMasks(0))

    numFilts = dssFilters(dssComp)
    do filtNum = 0, numFilts - 1
      dssFilt = dssFilter(dssComp, filtNum)
!print *, '  filtNum', filtNum
!print *, 'filter axis name:', dssFilterName(dssFilt)
      tmpDssFiltName = dssFilterName(dssFilt)
    if (trim(tmpDssFiltName) .ne. trim(axisName)) cycle
    !if (dssFilterName(dssFilt) .ne. axisName) cycle

      call getBitMasksDssFilt(dssFilt, filtBitMask)
!print *, '  size(filtBitMask):', size(filtBitMask)
!print *, '  filtBitMask%onBits:', filtBitMask%onBits
!print *, '  filtBitMask%offBits:', filtBitMask%offBits
      call andBitMasks(filtBitMask, totalBitMasks)
      deallocate(filtBitMask)
    end do ! filtNum
!print *, 'zzz getBitMasksDssComp'

  end subroutine getBitMasksDssComp

  !---------------------------------------------------------------------
  subroutine getBitMasksDssFilt(dssFilt, filtBitMask)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate filtBitMask after use.

    implicit none

    type(DSfilterT), intent(in) :: dssFilt
    type(BitMaskT),  pointer    :: filtBitMask(:)	! intent(out)

    integer :: filtType, nSize
    integer(int32), pointer :: onBitMasks(:) => null(), offBitMasks(:) => null()
    type(RangeT), allocatable :: ranges(:) 

    !-------------------------------------------------------------------
    ! Start

!print *, '    aaa getBitMasksDssFilt'

    filtType = dssFilterType(dssFilt)
    select case(filtType)
    case(BITMASKFILTERT)
      call dssFilterMask(dssFilt, onBitMasks, offBitMasks)
      if (size(onBitMasks) .ne. size(offBitMasks))&
call error('dssBitMaskBadSizes',"ssclib:getBitMasksDssFilt: Re the vectors &
&of bit masks returned by dssFilterMask, the the size of the vector of on-bits &
&is not equal to the size of the vector of off-bits.")

      allocate(filtBitMask(size(onBitMasks)))
      filtBitMask%onBits  = onBitMasks
      filtBitMask%offBits = offBitMasks
!print *, '    size(onBitMasks):', size(onBitMasks)
!print *, '    onBitMasks:', onBitMasks
!print *, '    offBitMasks:', offBitMasks

!***      deallocate(onBitMasks, offBitMasks)

      if (bitMaskFilterIsBad(filtBitMask))&
call error('dssBitMaskBadForm',"ssclib:getBitMasksDssFilt: Re the vectors &
&of bit masks returned by dssFilterMask, either the vector is of zero size &
&or, for at least 1 bit pair of 1 vector element, both on- and off-bit are set &
&(which is a nonsense).")

    case(RANGEFILTERT)
      call dssFilterRangesSize(dssFilt,nSize)
      allocate(ranges(0:nSize-1))
      ranges = dssRanges(dssFilt)
      call convertRangeFilterToBitMask(ranges, filtBitMask)
      deallocate(ranges)

    case default
      write(errStr, '(a, i2.2, a, i2.2, a, i2.2, a)') "ssclib:getBitMasksDssFilt: &
        &The input filter, which is of type ID ", filtType, ", is not recognized - &
        &it must be either BITMASKFILTERT (type ID ", BITMASKFILTERT&
        , ") or RANGEFILTERT (type ID ", RANGEFILTERT, ")."
call error('dssBitMaskBadFilterType')
    end select

!print *, '    zzz getBitMasksDssFilt'

  end subroutine getBitMasksDssFilt

  !---------------------------------------------------------------------
  subroutine initializeBitMask(bitMasks)
  !---------------------------------------------------------------------

    implicit none

    type(BitMaskT), intent(out) :: bitMasks(:)

    !-------------------------------------------------------------------
    ! Start

    bitMasks%onBits  = 0
    bitMasks%offBits = 0

  end subroutine initializeBitMask

  !---------------------------------------------------------------------
  logical(bool) function allBitsEquiv(bitMasks, bitNum)
  !---------------------------------------------------------------------

    implicit none

    type(BitMaskT), intent(in) :: bitMasks(:)
    integer(int32), intent(in) :: bitNum

    !-------------------------------------------------------------------
    ! Start

    
    if ((      all(btest(bitMasks%onBits,  bitNum))&
    .or. .not. any(btest(bitMasks%onBits,  bitNum)))&
    .and. (    all(btest(bitMasks%offBits, bitNum))&
    .or. .not. any(btest(bitMasks%offBits, bitNum))))&
    then

      allBitsEquiv = .true.
    else
      allBitsEquiv = .false.
    end if

  end function allBitsEquiv

  !---------------------------------------------------------------------
  subroutine convertRangeFilterToBitMask(ranges, bitMasks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate bitMasks after use.

    implicit none

    type(RangeT),   intent(in) :: ranges(:)
    type(BitMaskT), pointer    :: bitMasks(:)

    integer :: maxBlockN, blockN
    integer(int32) ::&
      numRanges, ri, maxNumMaskElements, q, maxBlockLength,&
      numMaskElementsThisSequence, maskNum, blockStartI, lengthRemainder,&
      blockLength, blockBitMask
    type(BitMaskT), allocatable :: tempBitMasks(:)
    integer(int32), allocatable :: rangeLoI(:), rangeHiI(:)

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(ranges)
    if (numRanges <= 0) then
      allocate(bitMasks(numRanges))
  return
    end if

    if (.not. rangesAreWellFormed(ranges))&
call error('badRangesForm',"ssclib:convertRangeFilterToBitMask: The input &
&ranges are not 'well-formed'")

    ! First step is to convert the range bounds to int32 values:
    allocate(rangeLoI(size(ranges)), rangeHiI(size(ranges)))
    do ri = 1, numRanges
      rangeLoI(ri) = lowestInt32InRange(ranges(ri))
      rangeHiI(ri) = highestInt32InRange(ranges(ri))
    end do

    ! Now we want to make some estimate of the maximum possible number of
    ! entries in the output pointer bitMasks. If we can estimate this number
    ! than we can allocate a temporary bit mask vector to this size and
    ! load the elements of this one by one. Then afterwards we can count
    ! how many elements we've actually got and at that point allocate the
    ! output vector.
    !
    ! The size of the output vector depends a bit on how clever we are. A
    ! brute-force approach would be to assign 1 element for each integer
    ! which passed the ranges. However this could be a lot of integers. In
    ! this routine we will be much cleverer because we note that any
    ! sequence {I_1,.. I_q} of integers which passes a single range can be
    ! broken up into blocks of length l=2**n, each of which has the same
    ! pattern in the bits b from b=msb to b=n (taking the least significant
    ! bit as bit 0 not 1). Such a block can be assigned a single element
    ! of bitMasks: this element will have onBits set to 1 for each 1 in
    ! the bit range b_msb to n+1, offBits set to 1 for each 0 in the bit
    ! range b_msb to n+1, and both onBits and offBits set to 0 otherwise.
    ! For example, a range which passed integers 7 to 13 passes therefore
    ! the following bit patterns:
    !
    !	Integer:	Bit pattern:
    !	-----------------------------
    !	 7		0111
    !	 8		1000
    !	 9		1001
    !	10		1010
    !	11		1011
    !	12		1100
    !	13		1101
    !	-----------------------------
    !
    ! The integers 8 to 11 form a block of length 2**2=4. This block can
    ! be represented by the following single bitMask element:
    !
    !	bitMask%onBits  = (..000)1000
    !	bitMask%offBits = (..111)0100
    !
    ! In other words, the first bit must be set, the second bit must be
    ! unset, and the remainder can have any value.
    !
    ! The largest such block of integers in a sequence of length q has n
    ! given by n=log_2(q+1)-1. (Note that one might be able to carve a
    ! block twice as big as this out of the sequence, depending on where
    ! it starts; however the number given is the maximum for any sequence
    ! start point.) For the example above, the length q=7 gives n=2
    ! thus l=4 as expected. Now, how to estimate the maximum total
    ! number of blocks in the sequence (hence maximum number of maskBits
    ! elements needed)? I don't have a proof that the number I calculate
    ! is an upper bound, but consider the following hand-waving argument.
    ! Suppose the sequence occurs such that a block of the largest size
    ! l_max falls as near to the middle of the sequence as possible. (The
    ! example above happens to satisfy this condition.) The most numerous
    ! arrangement of blocks in the remainders on either side will occur
    ! when the block sizes decrease towards the edge of the sequence as
    ! 2**(n_max-1), 2**(n_max-2) and so forth. (Note that the block sizes
    ! can never _increase_ from the middle towards the edges.) Suppose
    ! the remainder on one side is of length l_rem such that
    !
    !	l_rem = (q-l_max)/2 = (2**(n_max+1))-1.
    !
    ! We find in this case that we can fit n_max+1 blocks into l_rem. This
    ! seems to be the maximal case. Let us therefore calculate
    ! m = ceiling(log_2(l_rem+1)) as the maximum number of blocks that can
    ! be fitted into a generic remainder of length l_rem; the total upper
    ! bound M for the number of blocks in the entire sequence of length q
    ! is then
    !
    !	M = 1 + 2*(log_2[l_rem+1])
    !	  = 1 + 2*(log_2[{(q-l_max)/2}+1])
    !	  = 1 + 2*(log_2[{(q-2**n_max)/2}+1])
    !	  = 1 + 2*(log_2[{(q-2**[log_2(q+1)-1])/2}+1]).

    maxNumMaskElements = 0
    do ri = 1, numRanges
      q = 1 + rangeHiI(ri) - rangeLoI(ri)
    if (q <= 0) cycle

      maxBlockN = calcMostSignifSetBit(q + 1_int32) - 1 ! in ssclib/ssc_misc
      maxBlockLength = ibset(0_int32, maxBlockN)
      numMaskElementsThisSequence&
        = 1 + 2 * ceiling(log(ceiling((q - maxBlockLength) / 2.0) + 1.0)&
        / lnTwo)
      maxNumMaskElements = maxNumMaskElements + numMaskElementsThisSequence
    end do

    allocate(tempBitMasks(maxNumMaskElements))
    call initializeBitMask(tempBitMasks) ! sets everything to zero.

    ! Now we do the actual conversion:

    maskNum = 0
    do ri = 1, numRanges
      blockStartI = rangeLoI(ri)
      do while(blockStartI <= rangeHiI(ri))
        lengthRemainder = 1 + rangeHiI(ri) - blockStartI
        blockN = calcMostSignifSetBit(lengthRemainder)
        blockN = min(blockN, calcLeastSignifSetBit(blockStartI))
        ! these function calls are in ssclib/ssc_misc.
        blockLength = ibset(0_int32, blockN)

        ! In the case of the example given above, when blockStartI=7,
        ! right at the start, lengthRemainder=7, ie the whole length of
        ! the sequence lies ahead of us. However we cannot get a block
        ! larger than 1 out from this starting value because the
        ! maxBlockNth bit changes in going from blockStartI to
        ! blockStartI+1. Hence the 'min' statement. The next block will
        ! occur at blockStartI=8. In this case lengthRemainder=6.
        ! LSB(blockStartI)=3 but MSB(lengthRemainder)=2, hence the block
        ! is of length 2**2=4. The values for the final block are
        !
        !	blockStartI=12 
        !	lengthRemainder=2 
        !	LSB(blockStartI)=2 
        !	MSB(lengthRemainder)=1
        !
        ! Hence the final block is of length 2**1=2. 

        blockBitMask = not(blockLength-1) ! so, for the example given
        ! before, if blockStartI=8, we are in the process of assigning the
        ! central block of four integers 8 to 11 to a single bitMask. In
        ! this case blockLength=4 and the bit pattern of blockBitMask is
        ! 1100. We will now AND this blockBitMask with any of the integers
        ! in the block to produce the onBits 1000; then AND blockBitMask
        ! with NOT of any of the integers in the block to produce the
        ! offBits 0100.

        maskNum = maskNum + 1
        tempBitMasks(maskNum)%onBits  = iand(blockBitMask, blockStartI)
        tempBitMasks(maskNum)%offBits = iand(blockBitMask, not(blockStartI))

        blockStartI = blockStartI + blockLength
      end do
    end do

    allocate(bitMasks(maskNum))
    bitMasks = tempBitMasks(:maskNum)

  end subroutine convertRangeFilterToBitMask

  !---------------------------------------------------------------------
  logical(bool) function bitMaskFilterIsBad(flagFilter)
  !---------------------------------------------------------------------

    implicit none

    type(BitMaskT), intent(in) :: flagFilter(:) ! typedef in ssclib/dss_bitmasks_aux.

    !-------------------------------------------------------------------
    ! Start

    bitMaskFilterIsBad = .true. ! default
  if (size(flagFilter) <= 0) return
  if (any(iand(flagFilter%onBits, flagFilter%offBits) > 0)) return
    bitMaskFilterIsBad = .false. ! passed its tests

  end function bitMaskFilterIsBad

  !---------------------------------------------------------------------
  integer(int32) function lowestInt32InRange(range)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: range

    !-------------------------------------------------------------------
    ! Start

    if (range%lower%type == UNDEFINED) then
      lowestInt32InRange = ibset(0_int32, 31)
    else if (range%lower%type == INCLUSIVE) then
      lowestInt32InRange = ceiling(range%lower%value)
    else ! range%lower%type == EXCLUSIVE
      if (range%lower%value == nint(range%lower%value)) then
        lowestInt32InRange = nint(range%lower%value) + 1
      else
        lowestInt32InRange = ceiling(range%lower%value)
      end if
    end if

  end function lowestInt32InRange

  !---------------------------------------------------------------------
  integer(int32) function highestInt32InRange(range)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: range

    !-------------------------------------------------------------------
    ! Start

    if (range%upper%type == UNDEFINED) then
!      highestInt32InRange = not(ibset(0_int32, 31))
highestInt32InRange = huge(1_int32) !*** which is better??
    else if (range%upper%type == INCLUSIVE) then
      highestInt32InRange = floor(range%upper%value)
    else ! range%upper%type == EXCLUSIVE
      if (range%upper%value == nint(range%upper%value)) then
        highestInt32InRange = nint(range%upper%value) - 1
      else
        highestInt32InRange = floor(range%upper%value)
      end if
    end if

  end function highestInt32InRange

  !---------------------------------------------------------------------
  subroutine andBitMasksScalVec(inBitMask, andedBitMasks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedBitMasks after use.

    implicit none

    type(BitMaskT), intent(in) :: inBitMask
    type(BitMaskT), pointer    :: andedBitMasks(:)

    type(BitMaskT), pointer :: tempBitMasks(:)

    !-------------------------------------------------------------------
    ! Start

    call andBitMasksVecVecVec((/inBitMask/), andedBitMasks, tempBitMasks)
    deallocate(andedBitMasks)
!    allocate(andedBitMasks(size(tempBitMasks)))
!    andedBitMasks = tempBitMasks
!    dealocate(tempBitMasks)
andedBitMasks => tempBitMasks	! tempBitMasks should not be deallocated here.

  end subroutine andBitMasksScalVec

  !---------------------------------------------------------------------
  subroutine andBitMasksScalScalVec(bitMaskA, bitMaskB, andedBitMasks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedBitMasks after use.

    implicit none

    type(BitMaskT), intent(in) :: bitMaskA, bitMaskB
    type(BitMaskT), pointer    :: andedBitMasks(:)

    !-------------------------------------------------------------------
    ! Start

    call andBitMasksVecVecVec((/bitMaskA/), (/bitMaskB/), andedBitMasks)

  end subroutine andBitMasksScalScalVec

  !---------------------------------------------------------------------
  subroutine andBitMasksScalVecVec(bitMaskA, bitMasksB, andedBitMasks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedBitMasks after use.

    implicit none

    type(BitMaskT), intent(in) :: bitMaskA, bitMasksB(:)
    type(BitMaskT), pointer    :: andedBitMasks(:)

    !-------------------------------------------------------------------
    ! Start

    call andBitMasksVecVecVec((/bitMaskA/), bitMasksB, andedBitMasks)

  end subroutine andBitMasksScalVecVec

  !---------------------------------------------------------------------
  subroutine andBitMasksVecScalVec(bitMasksA, bitMaskB, andedBitMasks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedBitMasks after use.

    implicit none

    type(BitMaskT), intent(in) :: bitMasksA(:), bitMaskB
    type(BitMaskT), pointer    :: andedBitMasks(:)

    !-------------------------------------------------------------------
    ! Start

    call andBitMasksVecVecVec(bitMasksA, (/bitMaskB/), andedBitMasks)

  end subroutine andBitMasksVecScalVec

  !---------------------------------------------------------------------
  subroutine andBitMasksVecVec(inBitMasks, andedBitMasks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedBitMasks after use.

    implicit none

    type(BitMaskT), intent(in) :: inBitMasks(:)
    type(BitMaskT), pointer    :: andedBitMasks(:)

    type(BitMaskT), pointer :: tempBitMasks(:)

    !-------------------------------------------------------------------
    ! Start

    call andBitMasksVecVecVec(inBitMasks, andedBitMasks, tempBitMasks)

    deallocate(andedBitMasks)
!    allocate(andedBitMasks(size(tempBitMasks)))
!    andedBitMasks = tempBitMasks
!    dealocate(tempBitMasks)
andedBitMasks => tempBitMasks	! tempBitMasks should not be deallocated here.

  end subroutine andBitMasksVecVec

  !---------------------------------------------------------------------
  subroutine andBitMasksVecVecVec(bitMasksA, bitMasksB, andedBitMasks)
  !---------------------------------------------------------------------
  ! This ANDing is a bit tricky. Recall that the BitMaskT structure consists of an onBits and an offBits integer. Each bit position may have onBits set or offBits set or neither. Both on- and offBits set is not allowed. These three allowed states for the ith bit really mean, respectively, 'the ith bit of the filtered integer must be (i) set, (ii) unset, (iii) have either state,' ie in shorthand the required bit value is 0, 1 or X. ANDing two bitmasks proceeds according to the bit-wise truth table given in andBitMasksScalScal.
  !
  ! Let us look at an example. To make things easy suppose we only have 4 bits per integer instead of 32. Suppose we have two bitmask filters to AND together, each of which has two elements. (Recall that the elements of a bitmask are assumed to be related by a logical OR.) Let these be as follows:
  !
  !		bmA(1) = 10X0
  !		bmA(2) = X001
  !		bmB(1) = 0X10
  !		bmB(2) = 111X
  !
  ! Firstly, the result of the operation must be a bitmask vector with 2 x 2 = 4 elements. Why is this? Because the individual bit filters for any given element are assumed to be ANDed together. In other words bmA(1), when used to define the state of an integer Q, translates as 'the most significant bit of Q must be set, AND the next bit of Q must be unset, AND the next bit of Q can have any value, AND the least significant bit of Q must be unset.' So the result of the AND between bmA and bmB must have 4 elements which are evaluated as follows:
  !
  !		result(1) = bmA(1) AND bmB(1)
  !		result(2) = bmA(1) AND bmB(2)
  !		result(3) = bmA(2) AND bmB(1)
  !		result(4) = bmA(2) AND bmB(2)
  !
  ! According to the truth table above, the bit patterns of the elements of the result will therefore be:
  !
  !		result(1) = 00X0
  !		result(2) = 10X0
  !		result(3) = 0000
  !		result(4) = X00X
  !
  !---------------------------------------------------------------------

    implicit none

    type(BitMaskT), intent(in) :: bitMasksA(:), bitMasksB(:)
    type(BitMaskT), pointer    :: andedBitMasks(:)

    integer :: numElementsA, numElementsB, numElementsResult, ai, bi, ri

    !-------------------------------------------------------------------
    ! Start

    numElementsA = size(bitMasksA)
    numElementsB = size(bitMasksB)

    if (numElementsA <= 0) then
      if (numElementsB <= 0) then
        allocate(andedBitMasks(0))
      else
        allocate(andedBitMasks(numElementsB))
        andedBitMasks = bitMasksB
      end if
  return
    end if

    if (numElementsB <= 0) then
      allocate(andedBitMasks(numElementsA))
      andedBitMasks = bitMasksA
  return
    end if

    numElementsResult = numElementsA * numElementsB
    allocate(andedBitMasks(numElementsResult))
    do ai = 1, numElementsA
      do bi = 1, numElementsB
        ri = numElementsB * (ai - 1) + bi
        call andBitMasksScalScalScal(bitMasksA(ai), bitMasksB(bi)&
          , andedBitMasks(ri))
      end do
    end do

  end subroutine andBitMasksVecVecVec

  !---------------------------------------------------------------------
  subroutine andBitMasksScalScal(inBitMask, andedBitMask)
  !---------------------------------------------------------------------

    implicit none

    type(BitMaskT), intent(in)  :: inBitMask
    type(BitMaskT), intent(out) :: andedBitMask

    type(BitMaskT) :: tempBitMask

    !-------------------------------------------------------------------
    ! Start

    call andBitMasksScalScalScal(inBitMask, andedBitMask, tempBitMask)
    andedBitMask = tempBitMask

  end subroutine andBitMasksScalScal

  !---------------------------------------------------------------------
  subroutine andBitMasksScalScalScal(bitMaskA, bitMaskB, andedBitMask)
  !---------------------------------------------------------------------
  ! This ANDing is a bit tricky. Recall that the BitMaskT structure consists of an onBits and an offBits integer. Each bit position may have onBits set or offBits set or neither. Both on- and offBits set is not allowed. These three allowed states for the ith bit really mean, respectively, 'the ith bit of the filtered integer must be (i) set, (ii) unset, (iii) have either state,' ie in shorthand the required bit value is 0, 1 or X. ANDing two bitmasks must therefore proceed according to the following bit-wise truth table:
  !
  !		bitmask A	bitmask B	A AND B
  !		---------------------------------------
  !		   0		   0		   0
  !		   0		   1		   0
  !		   0		   X		   0
  !		   1		   0		   0
  !		   1		   1		   1
  !		   1		   X		   X
  !		   X		   0		   0
  !		   X		   1		   X
  !		   X		   X		   X
  !		---------------------------------------
  !
  !---------------------------------------------------------------------

    implicit none

    type(BitMaskT), intent(in)  :: bitMaskA, bitMaskB
    type(BitMaskT), intent(out) :: andedBitMask

    integer :: bit

    !-------------------------------------------------------------------
    ! Start

    do bit = 0, bit_size(bitMaskA%onBits)-1
      if (btest(bitMaskA%offBits, bit)) then
        if (btest(bitMaskB%offBits, bit)) then		! 0 0
          andedBitMask%offBits = ibset(andedBitMask%offBits, bit)	! 0
          andedBitMask%onBits  = ibclr(andedBitMask%onBits,  bit)
        else if (btest(bitMaskB%onBits, bit)) then	! 0 1
          andedBitMask%offBits = ibset(andedBitMask%offBits, bit)	! 0
          andedBitMask%onBits  = ibclr(andedBitMask%onBits,  bit)
        else						! 0 X
          andedBitMask%offBits = ibset(andedBitMask%offBits, bit)	! 0
          andedBitMask%onBits  = ibclr(andedBitMask%onBits,  bit)
        end if
      else if (btest(bitMaskA%onBits, bit)) then
        if (btest(bitMaskB%offBits, bit)) then		! 1 0
          andedBitMask%offBits = ibset(andedBitMask%offBits, bit)	! 0
          andedBitMask%onBits  = ibclr(andedBitMask%onBits,  bit)
        else if (btest(bitMaskB%onBits, bit)) then	! 1 1
          andedBitMask%offBits = ibclr(andedBitMask%offBits, bit)	! 1
          andedBitMask%onBits  = ibset(andedBitMask%onBits,  bit)
        else						! 1 X
          andedBitMask%offBits = ibclr(andedBitMask%offBits, bit)	! X
          andedBitMask%onBits  = ibclr(andedBitMask%onBits,  bit)
        end if
      else
        if (btest(bitMaskB%offBits, bit)) then		! X 0
          andedBitMask%offBits = ibset(andedBitMask%offBits, bit)	! 0
          andedBitMask%onBits  = ibclr(andedBitMask%onBits,  bit)
        else if (btest(bitMaskB%onBits, bit)) then	! X 1
          andedBitMask%offBits = ibclr(andedBitMask%offBits, bit)	! X
          andedBitMask%onBits  = ibclr(andedBitMask%onBits,  bit)
        else						! X X
          andedBitMask%offBits = ibclr(andedBitMask%offBits, bit)	! X
          andedBitMask%onBits  = ibclr(andedBitMask%onBits,  bit)
        end if
      end if
    end do

  end subroutine andBitMasksScalScalScal

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module dss_bitmasks_aux
