! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: fftw3_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains a header for fftw (version 3) and one or two other helpful routines. NOTE that a program should NOT use both this module and fftw2_aux - they are incompatible.
! 
! Comments: TODO:
! 
! $Id: fftw3_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module fftw3_aux

  use types
  use errorHandling

  use constants

  implicit none 

  public

!XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
!     This file contains PARAMETER statements for various constants
!     that can be passed to FFTW routines.  You should include
!     this file in any FORTRAN program that calls the sfftw
!     routines (either directly or with an #include statement
!     if you use the C preprocessor).

      integer, parameter, public ::&
        FFTW_FORWARD          =  -1,&
        FFTW_MEASURE          =   0,&
        FFTW_BACKWARD         =  +1,&
!        FFTW_REAL_TO_COMPLEX  =  -1,&
!        FFTW_COMPLEX_TO_REAL  =   1,&
        FFTW_EXHAUSTIVE       =   8,&
        FFTW_PATIENT          =  32,&
        FFTW_ESTIMATE         =  64,&
        FFTW_ESTIMATE_PATIENT = 128
!        FFTW_OUT_OF_PLACE     =   0,&
!        FFTW_IN_PLACE         =   8,&
!        FFTW_USE_WISDOM       =  16,&
!        FFTW_THREADSAFE       = 128

!!     Constants for the MPI wrappers:
!      integer, parameter, public ::&
!        FFTW_TRANSPOSED_ORDER =     1,&
!        FFTW_NORMAL_ORDER     =     0,&
!        FFTW_SCRAMBLED_INPUT  =  8192,&
!        FFTW_SCRAMBLED_OUTPUT = 16384

!XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  ! Arrangements so that fftw calls work on all platforms:
  type FftwPlanType
    integer, pointer :: cPointer
  end type

  private template
  private calcPowerSpectrum
  public findNextHighest2357multiple
  public symmetriseAryForFFT
  public calcFftwRedundantXSize
  public unfoldFftwRedundantX
  public cyclicConvolutionFft		! See also array_utils

  interface unfoldFftwRedundantX
    module procedure unfoldFftwRedundantX1d, unfoldFftwRedundantX2d
  end interface

  interface cyclicConvolutionFft	! See also array_utils
    module procedure cyclicConvFftSingle1d, cyclicConvFftSingle2d
  end interface

contains 

  !---------------------------------------------------------------------
  function findNextHighest2357multiple(i) result(result)
  !---------------------------------------------------------------------
  ! This is designed to look for the smallest integer that satisfies the
  ! following conditions: (i) it is greater than or equal to the argument
  ! i; (ii) it is a product only of the numbers 2, 3, 5 and 7. The library
  ! fftw works most efficiently on arrays that have the second property,
  ! so images are padded in size (i) to prevent any 'wrapping' of the
  ! convolver and (ii) so they have dimensions divisable by 2, 3, 5 and/or 7.

    implicit none

    integer, intent(in) ::&
      i
    integer ::&
      result

    integer ::&
      maxa,&
      maxb,&
      maxc,&
      maxd,&
      a,&
      b,&
      c,&
      d,&
      besta,&
      bestb,&
      bestc,&
      bestd
    real(single) ::&
      lni,&
      term,&
      bestTerm
    logical ::&
      firstOver

    !-------------------------------------------------------------------
    ! Start:

    lni = log(real(i))

    maxa = ceiling(lni / lnTwo)		! These are defined
    maxb = ceiling(lni / lnThree)	! in ssclib/constants.
    maxc = ceiling(lni / lnFive)	!
    maxd = ceiling(lni / lnSeven)	!

    firstOver = .true.
    bestTerm = 0.0 ! just to define it, so the inner 'if' generates no
    ! complaints from the compiler.
    do a = 0, maxa
      do b = 0, maxb
        do c = 0, maxc
          DLOOP: do d = 0, maxd
            term = a * lnTwo + b * lnThree + c * lnFive + d * lnSeven
            if (term >= lni .and. (term < bestTerm .or. firstOver)) then
              bestTerm = term
              besta = a
              bestb = b
              bestc = c
              bestd = d

              if (firstOver) firstOver = .false.

          exit DLOOP
            end if
          end do DLOOP
        end do
      end do
    end do

    result = (2**besta) * (3**bestb) * (5**bestc) * (7**bestd)

  end function findNextHighest2357multiple

  !---------------------------------------------------------------------
  subroutine calcPowerSpectrum(fftArray, psArray)
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(in) :: fftArray(:, :)
    real(single), intent(out) :: psArray(:, :)

    !-------------------------------------------------------------------
    ! Start

    psArray = fftArray * conjg(fftArray)

  end subroutine calcPowerSpectrum

  !---------------------------------------------------------------------
  subroutine symmetriseAryForFFT(array)
  !---------------------------------------------------------------------
  ! The purpose of this subroutine is manipulate a real array (of size (I, J)) which has the property that
  !
  !	array(i,j) = array(i,J+2-j) = array(I+2-i,j) = array(I+2-i,J+2-j)
  !
  ! for i and j > 1. In other words, it should be symmetrical about reflection in either the i = I/2+1 or j = J/2+1 axes, and therefore give rise to no imaginary terms upon back-fft.
  !
  ! The object is accomplished by averaging the contributions from the four corners of the array.

    implicit none

    real(single), intent(inout) :: array(:, :)

    integer(int16) :: xSize, ySize, xi, yi

    !-------------------------------------------------------------------
    ! Start

    xSize = size(array, 1)
    ySize = size(array, 2)

    do xi = 2, xSize/2 + 1
      do yi = 2, ySize/2 + 1
        array(xi, yi) = (array(xi, yi) + array(xi, ySize + 2 - yi)&
        + array(xSize + 2 - xi, yi) + array(xSize + 2 - xi, ySize + 2 - yi))&
        / 4.0
        array(xi,             ySize + 2 - yi) = array(xi, yi)
        array(xSize + 2 - xi, yi)             = array(xi, yi)
        array(xSize + 2 - xi, ySize + 2 - yi) = array(xi, yi)
      end do
    end do

  end subroutine symmetriseAryForFFT

  !---------------------------------------------------------------------
  integer function calcFftwRedundantXSize(xiSize)
  !---------------------------------------------------------------------
  ! This is not only used to calculate the x-coordinate dimension which
  ! fftw assigns to the non-redundant part of hermitian 2D arrays, it also
  ! offers a handy means of calculating the 'centre' of an array of
  ! arbitrary size. For example, whereas arrays suitable for FFT should
  ! have their 'central value' at the bottom left pixel (pixel (1,1) in
  ! fortran terms, (0,0) in c-ish), human beings prefer the central value
  ! to be at the central pixel. To convert from 'fft-natural' to 'human-
  ! natural' format, you could do as follows:
  !
  !	cxi = calcFftwRedundantXSize(size(fftArray,1))
  !	cyi = calcFftwRedundantXSize(size(fftArray,2))
  !	humanArray = cshift(cshift(fftArray, 1-cxi, 1), 1-cyi, 2)

    implicit none

    integer, intent(in) :: xiSize

    !-------------------------------------------------------------------
    ! Start

    calcFftwRedundantXSize = 1 + xiSize / 2 ! equiv to 1 + floor(xiSize/2.0).

  end function calcFftwRedundantXSize

  !---------------------------------------------------------------------
  subroutine unfoldFftwRedundantX1d(truncatedVec, fullVec)
  !---------------------------------------------------------------------
  ! The fftw real-to-complex and complex-to-real routines store the
  ! complex side of the transform in a vector about half the size of the
  ! real side. This is possible because of the Hermitian redundancy of
  ! the (complex) transform of real data. The present routine takes such
  ! a 'truncated' result of a real-to-complex fftw transform and unfolds
  ! it to the same size as the starting real-valued array. The result
  ! should be identical to what you would get if you first converted your
  ! real-valued array to complex and then performed a normal complex-to-
  ! complex fftw transform.
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(in)  :: truncatedVec(:)
    complex(single), intent(out) :: fullVec(:)

    integer :: xiSize, redundantXiSize

    !-------------------------------------------------------------------
    ! Start

    xiSize = size(fullVec)
    redundantXiSize = calcFftwRedundantXSize(xiSize)

    if (size(truncatedVec) .ne. redundantXiSize)&
call error('badXSize',"")

    fullVec(1:redundantXiSize) = truncatedVec
    fullVec(xiSize:xiSize+2-redundantXiSize:-1) = conjg(truncatedVec(2:redundantXiSize))

  end subroutine unfoldFftwRedundantX1d

  !---------------------------------------------------------------------
  subroutine unfoldFftwRedundantX2d(xTruncatedArray, fullArray)
  !---------------------------------------------------------------------
  ! The fftw real-to-complex and complex-to-real routines store the
  ! complex side of the transform in an array about half the size of the
  ! real side. This is possible because of the Hermitian redundancy of
  ! the (complex) transform of real data. The present routine takes such
  ! a 'truncated' result of a real-to-complex fftw transform and unfolds
  ! it to the same size as the starting real-valued array. The result
  ! should be identical to what you would get if you first converted your
  ! real-valued array to complex and then performed a normal complex-to-
  ! complex fftw transform.
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(in)  :: xTruncatedArray(:,:)
    complex(single), intent(out) :: fullArray(:,:)

    integer :: xiSize, yiSize, redundantXiSize

    !-------------------------------------------------------------------
    ! Start

    xiSize = size(fullArray,1)
    yiSize = size(fullArray,2)
    redundantXiSize = calcFftwRedundantXSize(xiSize)

    if (size(xTruncatedArray,1) .ne. redundantXiSize)&
call error('badXSize',"")
    if (size(xTruncatedArray,2) .ne. yiSize)&
call error('badYSize',"")

    fullArray(1:redundantXiSize,:) = xTruncatedArray
    fullArray(xiSize:xiSize+2-redundantXiSize:-1,yiSize:2:-1) = conjg(xTruncatedArray(2:,2:yiSize))
    fullArray(xiSize:xiSize+2-redundantXiSize:-1,1) = conjg(xTruncatedArray(2:,1))

  end subroutine unfoldFftwRedundantX2d

  !---------------------------------------------------------------------
  subroutine cyclicConvFftSingle2d(convolver, substrate)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)    :: convolver(:,:)
    real(single), intent(inout) :: substrate(size(convolver,1),size(convolver,2))

    integer ::xiSize, yiSize, halfXiSize
    real(single), allocatable :: fftIn(:,:)
    complex(single), allocatable :: fftOut(:,:), fftSubstrate(:,:)
    type(FftwPlanType) :: planTo, planFrom
    real(single) :: ftScale

    !-------------------------------------------------------------------
    ! Start

    xiSize = size(convolver,1)
    yiSize = size(convolver,2)
!    halfXiSize = userXiSize/2+1
    halfXiSize = calcFftwRedundantXSize(xiSize)
    ftScale = real(xiSize) * real(yiSize)

    allocate(fftIn(xiSize,yiSize))
    allocate(fftOut(halfXiSize,yiSize))
    allocate(fftSubstrate(halfXiSize,yiSize))

    call sfftw_plan_dft_r2c_2d(planTo,   xiSize, yiSize,&
      fftIn, fftOut, FFTW_MEASURE)
    call sfftw_plan_dft_c2r_2d(planFrom, xiSize, yiSize,&
      fftOut, fftIn, FFTW_MEASURE)

    fftIn = substrate
    call sfftw_execute(planTo)
    fftSubstrate = fftOut

    fftIn = convolver
    call sfftw_execute(planTo)
    fftOut = fftOut * fftSubstrate

    call sfftw_execute(planFrom)
    substrate = fftIn / ftScale

    call sfftw_destroy_plan(planTo)
    call sfftw_destroy_plan(planFrom)

  end subroutine cyclicConvFftSingle2d

  !---------------------------------------------------------------------
  subroutine cyclicConvFftSingle1d(convolver, substrate)
  !---------------------------------------------------------------------

    implicit none

    real(single),    intent(in)    :: convolver(:)
    real(single),    intent(inout) :: substrate(size(convolver))

    integer :: iSize, halfISize
    real(single), allocatable :: fftIn(:)
    complex(single), allocatable :: fftOut(:), fftSubstrate(:)
    type(FftwPlanType) :: planTo, planFrom
    real(single) :: ftScale

    !-------------------------------------------------------------------
    ! Start

    iSize = size(convolver)
!    halfISize = userXiSize/2+1
    halfISize = calcFftwRedundantXSize(iSize)
    ftScale = real(iSize)

    allocate(fftIn(iSize))
    allocate(fftOut(halfISize))
    allocate(fftSubstrate(halfISize))

    call sfftw_plan_dft_r2c_1d(planTo,   iSize, fftIn, fftOut, FFTW_MEASURE)
    call sfftw_plan_dft_c2r_1d(planFrom, iSize, fftOut, fftIn, FFTW_MEASURE)

    fftIn = substrate
    call sfftw_execute(planTo)
    fftSubstrate = fftOut

    fftIn = convolver
    call sfftw_execute(planTo)
    fftOut = fftOut * fftSubstrate

    call sfftw_execute(planFrom)
    substrate = fftIn / ftScale

    call sfftw_destroy_plan(planTo)
    call sfftw_destroy_plan(planFrom)

  end subroutine cyclicConvFftSingle1d

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module fftw3_aux
