! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: interpolations
!
! Authors:  Ian Stewart - SSC/LUX
!           Masaaki Sakano - SSC/LUX
!          
! Description: Contains routines to interpolate onto 1- and 2-d grids.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module interpolations

  use types

  use array_utils	! for getBinNumber
  use linear		! for solveLinearTriDiag.
  use ssc_misc		! for fracVal.

  implicit none 

  private 

  integer, parameter, public ::&
    INTERP_LINEAR=0, INTERP_SPLINE=1,&
    SPLINE_FIXED_1ST=0, SPLINE_FIXED_2ND=1

  public interpolate
  public interpolate2d

  public calcY2ForCubicSplines
  public cubicSplineInterpolate
  public bilinearInterpolateSscLib	! bilinearInterpolate in mssllib already
  public estimateGradsForBicubic
  public loadWeightsForBicubic
  public calcCoeffsForBicubic
  public bicubicInterpolate

  public interpolateInPolarCoord

  interface interpolate
    module procedure interpolateSingle, interpolateDouble
    module procedure interpolateSingComplex
    module procedure interpolateDoubComplex
    module procedure interpolateSingleVec, interpolateDoubleVec
    module procedure interpolateSingCmplxVec
    module procedure interpolateDoubCmplxVec
  end interface
  interface interpolate2d
    module procedure interpolate2dSingle, interpolate2dDouble
    module procedure interpolate2dSingComplex
    module procedure interpolate2dDoubComplex
    module procedure interpolate2dSingleVec, interpolate2dDoubleVec
    module procedure interpolate2dSingCmplxVec
    module procedure interpolate2dDoubCmplxVec
  end interface

  interface calcY2ForCubicSplines
    module procedure calcY2ForCubicSplinesSing
    module procedure calcY2ForCubicSplinesDble
  end interface
  interface cubicSplineInterpolate
    module procedure cubicSplineInterpSingle
    module procedure cubicSplineInterpDouble
  end interface
  interface bilinearInterpolateSscLib
    module procedure bilinearInterpSingle
    module procedure bilinearInterpDouble
    module procedure bilinearInterpSingCmplx
    module procedure bilinearInterpDbleCmplx
  end interface
  interface estimateGradsForBicubic
    module procedure estimateGradsForBicubicSing
    module procedure estimateGradsForBicubicDble
  end interface
!  interface loadWeightsForBicubic
!    module procedure loadWeightsForBicubicSing
!    module procedure loadWeightsForBicubicDble
!  end interface
  interface calcCoeffsForBicubic
    module procedure calcCoeffsForBicubicSing
    module procedure calcCoeffsForBicubicDble
  end interface
  interface bicubicInterpolate
    module procedure bicubicInterpSingle
    module procedure bicubicInterpDouble
  end interface

!  public bilinearSample
!  public bilinearValue
!  public bilinearSampleVariance
!  public bilinearValueVariance

!  interface bilinearValue
!    module procedure bilinearValueSingle, bilinearValueDouble
!    module procedure bilinearValueSingComplex
!    module procedure bilinearValueDoubComplex
!  end interface

contains 

  !---------------------------------------------------------------------
  subroutine interpolateSingle(ysOld, yNew, xNew, xsOld, style&
  , nullElement)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: ysOld(:)
    real(single), intent(out)           :: yNew
    real(single), intent(in)            :: xNew
    real(single), intent(in),  optional :: xsOld(size(ysOld))
    integer,      intent(in),  optional :: style
    logical,      intent(out), optional :: nullElement

    real(single) :: ysNew(1)
    logical :: nullElements(1)

    !-------------------------------------------------------------------
    ! Start

    call interpolateSingleVec(ysOld, ysNew, (/xNew/), xsOld, style&
      , nullElements)

    yNew = ysNew(1)
    if (present(nullElement)) nullElement = nullElements(1)

  end subroutine interpolateSingle

  !---------------------------------------------------------------------
  subroutine interpolateDouble(ysOld, yNew, xNew, xsOld, style&
  , nullElement)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)            :: ysOld(:)
    real(double), intent(out)           :: yNew
    real(double), intent(in)            :: xNew
    real(double), intent(in),  optional :: xsOld(size(ysOld))
    integer,      intent(in),  optional :: style
    logical,      intent(out), optional :: nullElement

    real(double) :: ysNew(1)
    logical :: nullElements(1)

    !-------------------------------------------------------------------
    ! Start

    call interpolateDoubleVec(ysOld, ysNew, (/xNew/), xsOld, style&
      , nullElements)

    yNew = ysNew(1)
    if (present(nullElement)) nullElement = nullElements(1)

  end subroutine interpolateDouble

  !---------------------------------------------------------------------
  subroutine interpolateSingComplex(ysOld, yNew, xNew, xsOld, style&
  , nullElement)
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(in)            :: ysOld(:)
    complex(single), intent(out)           :: yNew
    real(single),    intent(in)            :: xNew
    real(single),    intent(in),  optional :: xsOld(size(ysOld))
    integer,         intent(in),  optional :: style
    logical,         intent(out), optional :: nullElement

    complex(single) :: ysNew(1)
    logical :: nullElements(1)

    !-------------------------------------------------------------------
    ! Start

    call interpolateSingCmplxVec(ysOld, ysNew, (/xNew/), xsOld, style&
      , nullElements)

    yNew = ysNew(1)
    if (present(nullElement)) nullElement = nullElements(1)

  end subroutine interpolateSingComplex

  !---------------------------------------------------------------------
  subroutine interpolateDoubComplex(ysOld, yNew, xNew, xsOld, style&
  , nullElement)
  !---------------------------------------------------------------------

    implicit none

    complex(double), intent(in)            :: ysOld(:)
    complex(double), intent(out)           :: yNew
    real(double),    intent(in)            :: xNew
    real(double),    intent(in),  optional :: xsOld(size(ysOld))
    integer,         intent(in),  optional :: style
    logical,         intent(out), optional :: nullElement

    complex(double) :: ysNew(1)
    logical :: nullElements(1)

    !-------------------------------------------------------------------
    ! Start

    call interpolateDoubCmplxVec(ysOld, ysNew, (/xNew/), xsOld, style&
      , nullElements)

    yNew = ysNew(1)
    if (present(nullElement)) nullElement = nullElements(1)

  end subroutine interpolateDoubComplex

  !---------------------------------------------------------------------
  subroutine interpolateSingleVec(ysOld, ysNew, xsNew, xsOld, style&
  , nullElements)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: ysOld(:)
    real(single), intent(out)           :: ysNew(:)
    real(single), intent(in)            :: xsNew(size(ysNew))
    real(single), intent(in),  optional :: xsOld(size(ysOld))
    integer,      intent(in),  optional :: style
    logical,      intent(out), optional :: nullElements(size(ysNew))

    complex(double), allocatable :: ysNewDoubCmplx(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(ysNewDoubCmplx(size(ysNew)))

    if (present(xsOld)) then
      call interpolateDoubCmplxVec(cmplx(ysOld, 0.0, double)&
        , ysNewDoubCmplx, real(xsNew, double)&
        , real(xsOld, double), style, nullElements)
    else
      call interpolateDoubCmplxVec(cmplx(ysOld, 0.0, double)&
        , ysNewDoubCmplx, real(xsNew, double)&
        , style=style, nullElements=nullElements)
    end if

    ysNew = real(ysNewDoubCmplx, single)

  end subroutine interpolateSingleVec

  !---------------------------------------------------------------------
  subroutine interpolateDoubleVec(ysOld, ysNew, xsNew, xsOld, style&
  , nullElements)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)            :: ysOld(:)
    real(double), intent(out)           :: ysNew(:)
    real(double), intent(in)            :: xsNew(size(ysNew))
    real(double), intent(in),  optional :: xsOld(size(ysOld))
    integer,      intent(in),  optional :: style
    logical,      intent(out), optional :: nullElements(size(ysNew))

    complex(double), allocatable :: ysNewDoubCmplx(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(ysNewDoubCmplx(size(ysNew)))

    call interpolateDoubCmplxVec(cmplx(ysOld, 0.0d0, double)&
      , ysNewDoubCmplx, xsNew, xsOld, style, nullElements)

    ysNew = real(ysNewDoubCmplx, single)

  end subroutine interpolateDoubleVec

  !---------------------------------------------------------------------
  subroutine interpolateSingCmplxVec(ysOld, ysNew, xsNew, xsOld, style&
  , nullElements)
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(in)            :: ysOld(:)
    complex(single), intent(out)           :: ysNew(:)
    real(single),    intent(in)            :: xsNew(size(ysNew))
    real(single),    intent(in),  optional :: xsOld(size(ysOld))
    integer,         intent(in),  optional :: style
    logical,         intent(out), optional :: nullElements(size(ysNew))

    complex(double), allocatable :: ysNewDoubCmplx(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(ysNewDoubCmplx(size(ysNew)))
    if (present(xsOld)) then
      call interpolateDoubCmplxVec(cmplx(ysOld, kind=double)&
        , ysNewDoubCmplx, real(xsNew, double)&
        , real(xsOld, double), style, nullElements)
    else
      call interpolateDoubCmplxVec(cmplx(ysOld, kind=double)&
        , ysNewDoubCmplx, real(xsNew, double)&
        , style=style, nullElements=nullElements)
    end if
!AI NAG 5.2 porting
!kind=single changed to kind=double
    ysNew = cmplx(ysNewDoubCmplx, kind=double)

  end subroutine interpolateSingCmplxVec

  !---------------------------------------------------------------------
  subroutine interpolateDoubCmplxVec(ysOld, ysNew, xsNew, xsOld, style&
  , nullElements)
  !---------------------------------------------------------------------


!******* make appropriate variables 'save', as in 2d case. This can save a lot of time when the present routine is called multiple times from within a loop, with the same ysOld.



    implicit none

    complex(double), intent(in)            :: ysOld(:)
    complex(double), intent(out)           :: ysNew(:)
    real(double),    intent(in)            :: xsNew(size(ysNew))
    real(double),    intent(in),  optional :: xsOld(size(ysOld))
    integer,         intent(in),  optional :: style
    logical,         intent(out), optional :: nullElements(size(ysNew))

    integer :: xiSizeOld, xiSizeNew, xi, localStyle, xxi, i
    integer, allocatable :: binNumbers(:)
    logical, allocatable :: localNulls(:)
    real(double), allocatable :: localXsOld(:), y2sReal(:), y2sImag(:)
    real(double) :: yNewReal, yNewImag, x

    logical, save :: firstTime = .true.
    integer, save :: lastBinNum
logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa interpolateDoubCmplxVec'

    xiSizeOld = size(ysOld)
    xiSizeNew = size(ysNew)

    if (xiSizeOld < 2)&
call error('tooFewOldXVals',"interpolateDoubCmplxVec: There must be more &
&than 2 elements to argument ysOld")

    if (firstTime) then
      firstTime = .false.
      lastBinNum = 0
    end if

    allocate(binNumbers(xiSizeNew), localNulls(xiSizeNew))
    allocate(localXsOld(xiSizeOld))

    if (present(xsOld)) then
      if (any(xsOld(1:xiSizeOld-1) >= xsOld(2:xiSizeOld)))&
call error('badOrderOldXValues',"interpolateDoubCmplxVec: The elements of &
&xsOld must be in monotonically increasing order.")

      do xi = 1, xiSizeNew
        lastBinNum = getBinNumber(xsOld, xsNew(xi), lastBinNum)
        binNumbers(xi) = lastBinNum
      end do

      localXsOld = xsOld

    else ! assume xsOld = (/(i, i=0,xiSizeOld-1)/)
      where(xsNew == xsOld(xiSizeOld))
        binNumbers = xiSizeOld-1 ! ie, place values which fall right on
        ! the RH boundary into the last bin. This is for symmetry because
        ! values falling exactly on the LH boundary are placed in the 1st
        ! bin; only values < the LH value or > the RH value receive null
        ! bin numbers. Note that this is also the way that getBinNumber()
        ! behaves.
      elsewhere
        binNumbers = 1 + floor(xsNew)
      end where

      localXsOld = (/(real(i, single), i=0,xiSizeOld-1)/)
    end if

    localNulls = .false. ! default
    where(binNumbers < 1 .or. binNumbers >= xiSizeOld)&
      localNulls = .true.

    ysNew = (0.0d0,0.0d0) ! default

    localStyle = decideOption(style, INTERP_LINEAR)
    select case (localStyle)
    case(INTERP_LINEAR)
      do xi = 1, xiSizeNew
      if (localNulls(xi)) cycle
        xxi = binNumbers(xi)
        x = xsNew(xi)
        ysNew(xi) = ysOld(xxi) + (x - localXsOld(xxi)) * (ysOld(xxi+1) - ysOld(xxi))&
          / (localXsOld(xxi+1) - localXsOld(xxi))
      end do

    case(INTERP_SPLINE)
      allocate(y2sReal(xiSizeOld), y2sImag(xiSizeOld))
      call calcY2ForCubicSplinesDble(xsOld, real(ysOld, double), y2sReal)
      call calcY2ForCubicSplinesDble(xsOld, real(aimag(ysOld), double), y2sImag)
!****** NOTE that no boundary styles are passed to these two calls, so the defaults are assumed.

      do xi = 1, xiSizeNew
      if (localNulls(xi)) cycle
        xxi = binNumbers(xi)
        x = xsNew(xi)
        yNewReal = cubicSplineInterpDouble(localXsOld(xxi:xxi+1)&
          , real(ysOld(xxi:xxi+1), double), y2sReal(xxi:xxi+1), x)
        yNewImag = cubicSplineInterpDouble(localXsOld(xxi:xxi+1)&
          , real(aimag(ysOld(xxi:xxi+1)), double), y2sImag(xxi:xxi+1), x)
        ysNew(xi) = cmplx(yNewReal, yNewImag)
      end do

    case default
call error('badStyle',"The value of the style argument was not recognized.")
    end select

    if (present(nullElements))&
      nullElements = localNulls
if (testFlag) print *, 'zz interpolateDoubCmplxVec'

  end subroutine interpolateDoubCmplxVec

  !---------------------------------------------------------------------
  subroutine calcY2ForCubicSplinesSing(xs, ys, y2s, loBoundType, hiBoundType&
  , loBoundValue, hiBoundValue)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xs(:),&
                                          ys(size(xs))
    integer,      intent(in), optional :: loBoundType,&
                                          hiBoundType
    real(single), intent(in), optional :: loBoundValue,&
                                          hiBoundValue
    real(single), intent(out)          :: y2s(size(xs))

    real(double), allocatable :: y2sDble(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(y2sDble(size(xs)))

    call calcY2ForCubicSplinesDble(real(xs, double), real(ys, double)&
      , y2sDble, loBoundType, hiBoundType&
      , real(loBoundValue, double), real(hiBoundValue, double))

    y2s = real(y2sDble, single)

  end subroutine calcY2ForCubicSplinesSing

  !---------------------------------------------------------------------
  subroutine calcY2ForCubicSplinesDble(xs, ys, y2s, loBoundType, hiBoundType&
  , loBoundValue, hiBoundValue)
  !---------------------------------------------------------------------
!**** Was modified from the routine in ssclib/splines.f90. Status argument removed (replaced by calls to error()); order of other arguments changed; types and values made optional; type parameters defined now in the present module. IMS 2006-09-11.

  ! This is designed to work in conjunction with subroutine
  ! cubicSplineInterpolate() to provide a cubic-spline-type interpolation.
  ! The present subroutine, given an ordered series of unique x values
  ! plus corresponding y values, calculates the values at all x of the
  ! second derivatives (returned in y2) of the spline function which is
  ! continuous in its second derivative and which obeys the boundary
  ! conditions.
  !
  ! In terms of the variable names used, the following equation is solved:
  !
  !		triDiag * y2 = b.
  !
  ! This routine was written with the help of Numerical Recipes for Fortran.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)           :: xs(:),&
                                          ys(size(xs))
    integer,      intent(in), optional :: loBoundType,&
                                          hiBoundType
    real(double), intent(in), optional :: loBoundValue,&
                                          hiBoundValue
    real(double), intent(out)          :: y2s(size(xs))

    integer :: n, i, localLoBoundType, localHiBoundType
    real(double) :: localLoBoundValue, localHiBoundValue
    real(double), allocatable :: triDiag(:,:), b(:)

    !-------------------------------------------------------------------
    ! Start

    y2s = 0.0
    n = size(xs)
  if (n <= 0) return ! no point in going further.

    localLoBoundType = decideOption(loBoundType, SPLINE_FIXED_2ND)
    localHiBoundType = decideOption(hiBoundType, SPLINE_FIXED_2ND)
    localLoBoundValue = decideOption(loBoundValue, 0.0d0)
    localHiBoundValue = decideOption(hiBoundValue, 0.0d0)

    allocate(triDiag(-1:1, n), b(n))

    select case(localLoBoundType)
    case(SPLINE_FIXED_1ST) ! loBoundValue is taken to be the value of the
    ! first derivative at the lowest x sample.

      triDiag(0, 1) = (xs(2) - xs(1)) / 3.0
      triDiag(1, 1) = (xs(2) - xs(1)) / 6.0
      b(1) = (ys(2) - ys(1)) / (xs(2) - xs(1)) - localLoBoundValue

    case(SPLINE_FIXED_2ND) ! loBoundValue is taken to be the value of the
    ! second derivative at the lowest x sample.

      triDiag(0, 1) = 1.0
      triDiag(1, 1) = 0.0
      b(1) = localLoBoundValue

    case default
call error('badLoBoundType',"The value of the loBoundType argument was not recognized.")
    end select

    select case(localHiBoundType)
    case(SPLINE_FIXED_1ST) ! hiBoundType is taken to be the value of the
    ! first derivative at the highest x sample.

      triDiag(-1, n) = (xs(n) - xs(n-1)) / 6.0
      triDiag( 0, n) = (xs(n) - xs(n-1)) / 3.0
      b(n) = localHiBoundValue - (ys(n) - ys(n-1)) / (xs(n) - xs(n-1))

    case(SPLINE_FIXED_2ND) ! hiBoundType is taken to be the value of the
    ! second derivative at the highest x sample.

      triDiag(-1, n) = 0.0
      triDiag( 0, n) = 1.0
      b(n) = localHiBoundValue

    case default
call error('badHiBoundType',"The value of the hiBoundType argument was not recognized.")
    end select

    ! Fill up the middle rows of triDiag and b:

    do i = 2, n - 1
      triDiag(-1, i) = (xs( i ) - xs(i-1)) / 6.0
      triDiag( 0, i) = (xs(i+1) - xs(i-1)) / 3.0
      triDiag( 1, i) = (xs(i+1) - xs( i )) / 6.0

      b(i) = (ys(i+1) - ys( i )) / (xs(i+1) - xs( i ))&
           - (ys( i ) - ys(i-1)) / (xs( i ) - xs(i-1))
    end do

    triDiag(-1, 1) = 0.0 ! always.
    triDiag( 1, n) = 0.0 ! always.

    ! Now solve the system of equations:

    call solveLinearTriDiag(triDiag, b, y2s) ! in ssclib/linear.f90.

  end subroutine calcY2ForCubicSplinesDble

  !---------------------------------------------------------------------
  function cubicSplineInterpSingle(xs, ys, y2s, xIn) result(yOut)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: xs(2), ys(2), y2s(2), xIn
    real(single)             :: yOut

    !-------------------------------------------------------------------
    ! Start

    yOut = cubicSplineInterpDouble(real(xs, double), real(ys, double)&
      , real(y2s, double), real(xIn, double))
    ! implicit conversion to double.

  end function cubicSplineInterpSingle

  !---------------------------------------------------------------------
  function cubicSplineInterpDouble(xs, ys, y2s, xIn) result(yOut)
  !---------------------------------------------------------------------
!**** Was modified from the routine in ssclib/splines.f90. Status argument removed (replaced by calls to error()); full-length vector arguments replaced by 2-element arguments; turned into a function. IMS 2006-09-11.

  ! This is designed to work in conjunction with subroutine
  ! calcCubicSplines() to provide a cubic-spline-type interpolation.
  ! The present subroutine, given an ordered series of unique x values,
  ! plus the corresponding y values, plus the corresponding values of
  ! the second derivative y2, and finally plus a single x value xIn,
  ! performs the interpolation using the formula
  !
  !	y = a*y(i) + b*y(i+1) + c*y2(i) + d*y2(i+1)
  !
  ! where
  !
  !	     x(i+1) - xIn
  !	a = ---------------, b = 1 - a,
  !	     x(i+1) - x(i)
  !
  !	     1
  !	c = --- a*(a*a - 1)*(x(i+1) - x(i))^2
  !	     6
  ! and
  !	     1
  !	d = --- b*(b*b - 1)*(x(i+1) - x(i))^2
  !	     6
  !
  ! to return in the variable yOut the value of y at xIn.
  !
  ! Note that xIn must be >= x(1) and <= x(n).
  !
  ! This routine was written with the help of Numerical Recipes for Fortran.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: xs(2), ys(2), y2s(2), xIn
    real(double)             :: yOut

    real(double) :: a, b, denom, commonTerm

    !-------------------------------------------------------------------
    ! Start

    denom = xs(2) - xs(1)
    a = (xs(2) - xIn)  / denom
    b = (xIn   - xs(1)) / denom

    commonTerm = denom * denom / 6.0
    yOut = a * (ys(1) + (a * a - 1.0) * commonTerm * y2s(1))&
         + b * (ys(2) + (b * b - 1.0) * commonTerm * y2s(2))

  end function cubicSplineInterpDouble


  !---------------------------------------------------------------------
  subroutine interpolate2dSingle(zsOld, zNew, xNew, yNew&
  , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
  , nullElement)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: zsOld(:,:)
    real(single), intent(out)           :: zNew
    real(single), intent(in),  optional :: xNew, yNew
    real(single), intent(in),  optional :: xStartOld, yStartOld,&
                                           xFinisOld, yFinisOld
    integer,      intent(in),  optional :: style
    logical,      intent(in),  optional :: forceNewStart
    logical,      intent(out), optional :: nullElement

    real(single) :: zsNew(1,1)
    logical :: nullElements(1,1)

    !-------------------------------------------------------------------
    ! Start

    call interpolate2dSingleVec(zsOld, zsNew, (/xNew/), (/yNew/)&
      , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
      , nullElements)

    zNew = zsNew(1,1)
    if (present(nullElement)) nullElement = nullElements(1,1)

  end subroutine interpolate2dSingle

  !---------------------------------------------------------------------
  subroutine interpolate2dDouble(zsOld, zNew, xNew, yNew&
  , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
  , nullElement)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)            :: zsOld(:,:)
    real(double), intent(out)           :: zNew
    real(double), intent(in),  optional :: xNew, yNew
    real(double), intent(in),  optional :: xStartOld, yStartOld,&
                                           xFinisOld, yFinisOld
    integer,      intent(in),  optional :: style
    logical,      intent(in),  optional :: forceNewStart
    logical,      intent(out), optional :: nullElement

    real(double) :: zsNew(1,1)
    logical :: nullElements(1,1)

    !-------------------------------------------------------------------
    ! Start

    call interpolate2dDoubleVec(zsOld, zsNew, (/xNew/), (/yNew/)&
      , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
      , nullElements)

    zNew = zsNew(1,1)
    if (present(nullElement)) nullElement = nullElements(1,1)

  end subroutine interpolate2dDouble

  !---------------------------------------------------------------------
  subroutine interpolate2dSingComplex(zsOld, zNew, xNew, yNew&
  , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
  , nullElement)
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(in)            :: zsOld(:,:)
    complex(single), intent(out)           :: zNew
    real(single),    intent(in),  optional :: xNew, yNew
    real(single),    intent(in),  optional :: xStartOld, yStartOld,&
                                              xFinisOld, yFinisOld
    integer,         intent(in),  optional :: style
    logical,         intent(in),  optional :: forceNewStart
    logical,         intent(out), optional :: nullElement

    complex(single) :: zsNew(1,1)
    logical :: nullElements(1,1)

    !-------------------------------------------------------------------
    ! Start

    call interpolate2dSingCmplxVec(zsOld, zsNew, (/xNew/), (/yNew/)&
      , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
      , nullElements)

    zNew = zsNew(1,1)
    if (present(nullElement)) nullElement = nullElements(1,1)

  end subroutine interpolate2dSingComplex

  !---------------------------------------------------------------------
  subroutine interpolate2dDoubComplex(zsOld, zNew, xNew, yNew&
  , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
  , nullElement)
  !---------------------------------------------------------------------

    implicit none

    complex(double), intent(in)            :: zsOld(:,:)
    complex(double), intent(out)           :: zNew
    real(double),    intent(in),  optional :: xNew, yNew
    real(double),    intent(in),  optional :: xStartOld, yStartOld,&
                                              xFinisOld, yFinisOld
    integer,         intent(in),  optional :: style
    logical,         intent(in),  optional :: forceNewStart
    logical,         intent(out), optional :: nullElement

    complex(double) :: zsNew(1,1)
    logical :: nullElements(1,1)

    !-------------------------------------------------------------------
    ! Start

!    call interpolate2dDoubCmplxVec_20061116(zsOld, zsNew, (/xNew/), (/yNew/)&
!      , xStartOld, yStartOld, xFinisOld, yFinisOld, style, nullElements)
    call interpolate2dDoubCmplxVec(zsOld, zsNew, (/xNew/), (/yNew/)&
      , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
      , nullElements)

    zNew = zsNew(1,1)
    if (present(nullElement)) nullElement = nullElements(1,1)

  end subroutine interpolate2dDoubComplex

  !---------------------------------------------------------------------
  subroutine interpolate2dSingleVec(zsOld, zsNew, xsNew, ysNew&
  , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
  , nullElements)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: zsOld(:,:)
    real(single), intent(out)           :: zsNew(:,:)
    real(single), intent(in),  optional :: xsNew(size(zsNew,1)), ysNew(size(zsNew,2))
    real(single), intent(in),  optional :: xStartOld, yStartOld,&
                                           xFinisOld, yFinisOld
    integer,      intent(in),  optional :: style
    logical,      intent(in),  optional :: forceNewStart
    logical,      intent(out), optional :: nullElements(size(zsNew,1),size(zsNew,2))

    complex(double), allocatable :: zsNewDoubCmplx(:,:)
!    real(double) ::&
!      localXStartOld, localYStartOld, localXFinisOld, localYFinisOld
logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'interpolate2dSingleVec aaa'

    allocate(zsNewDoubCmplx(size(zsNew,1),size(zsNew,2)))

!    localXStartOld = decideOption(xStartOld, 0.0))
!    localYStartOld = decideOption(yStartOld, 0.0))
!    localXFinisOld = decideOption(xFinisOld, real(xiSizeOld-1,single)))
!    localYFinisOld = decideOption(yFinisOld, real(yiSizeOld-1,single)))
!    ! This is not very safe (because the same calculation is done within interpolate2dDoubCmplxVec under different circumstances), but I can't think how else to arrange the overloading

    if (  present(xStartOld) .and. present(yStartOld)&
    .and. present(xFinisOld) .and. present(yFinisOld)) then
      if (present(xsNew) .and. present(ysNew)) then
!        call interpolate2dDoubCmplxVec_20061116(cmplx(zsOld, 0.0, double)&
!          , zsNewDoubCmplx, real(xsNew, double), real(ysNew, double)&
!          , real(xStartOld, double), real(yStartOld, double)&
!          , real(xFinisOld, double), real(yFinisOld, double), style, nullElements)
        call interpolate2dDoubCmplxVec(cmplx(zsOld, 0.0, double)&
          , zsNewDoubCmplx, real(xsNew, double), real(ysNew, double)&
          , real(xStartOld, double), real(yStartOld, double)&
          , real(xFinisOld, double), real(yFinisOld, double), style, forceNewStart&
          , nullElements)

      else if (.not. (present(xsNew) .or. present(ysNew))) then
!        call interpolate2dDoubCmplxVec_20061116(cmplx(zsOld, 0.0, double)&
!          , zsNewDoubCmplx&
!          , xStartOld=real(xStartOld, double)&
!          , yStartOld=real(yStartOld, double)&
!          , xFinisOld=real(xFinisOld, double)&
!          , yFinisOld=real(yFinisOld, double)&
!          , style=style, nullElements=nullElements)
        call interpolate2dDoubCmplxVec(cmplx(zsOld, 0.0, double)&
          , zsNewDoubCmplx&
          , xStartOld=real(xStartOld, double)&
          , yStartOld=real(yStartOld, double)&
          , xFinisOld=real(xFinisOld, double)&
          , yFinisOld=real(yFinisOld, double)&
          , style=style, forceNewStart=forceNewStart, nullElements=nullElements)

      else
call error('badArgumentCombination',"You must call interpolate2d with &
&either both or the optional arguments xsNew and ysNew, or neither.")
      end if

    else if (.not.(present(xStartOld) .or. present(yStartOld)&
              .or. present(xFinisOld) .or. present(yFinisOld))) then
      if (present(xsNew) .and. present(ysNew)) then
!        call interpolate2dDoubCmplxVec_20061116(cmplx(zsOld, 0.0, double)&
!          , zsNewDoubCmplx, real(xsNew, double), real(ysNew, double)&
!          , style=style, nullElements=nullElements)
        call interpolate2dDoubCmplxVec(cmplx(zsOld, 0.0, double)&
          , zsNewDoubCmplx, real(xsNew, double), real(ysNew, double)&
          , style=style, forceNewStart=forceNewStart, nullElements=nullElements)

      else if (.not. (present(xsNew) .or. present(ysNew))) then
if (testFlag) print *, 'interpolate2dSingleVec bbb'

!        call interpolate2dDoubCmplxVec_20061116(cmplx(zsOld, 0.0, double)&
!          , zsNewDoubCmplx, style=style, nullElements=nullElements)
        call interpolate2dDoubCmplxVec(cmplx(zsOld, 0.0, double)&
          , zsNewDoubCmplx, style=style, forceNewStart=forceNewStart&
          , nullElements=nullElements)

      else
call error('badArgumentCombination',"You must call interpolate2d with &
&either both or the optional arguments xsNew and ysNew, or neither.")
      end if

    else
call error('badArgumentCombination',"You must call interpolate2d with &
&either all the optional arguments xStartOld, yStartOld, xFinisOld and &
&yFinisOld, or none.")
    end if
    zsNew = cmplx(zsNewDoubCmplx, kind=double)
!AI NAG 5.2 porting
!kind=single changed to kind=double

if (testFlag) print *, 'interpolate2dSingleVec zzz'

  end subroutine interpolate2dSingleVec

  !---------------------------------------------------------------------
  subroutine interpolate2dDoubleVec(zsOld, zsNew, xsNew, ysNew&
  , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
  , nullElements)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)            :: zsOld(:,:)
    real(double), intent(out)           :: zsNew(:,:)
    real(double), intent(in),  optional :: xsNew(size(zsNew,1)), ysNew(size(zsNew,2))
    real(double), intent(in),  optional :: xStartOld, yStartOld,&
                                           xFinisOld, yFinisOld
    integer,      intent(in),  optional :: style
    logical,      intent(in),  optional :: forceNewStart
    logical,      intent(out), optional :: nullElements(size(zsNew,1),size(zsNew,2))

    complex(double), allocatable :: zsNewDoubCmplx(:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(zsNewDoubCmplx(size(zsNew,1),size(zsNew,2)))

!    call interpolate2dDoubCmplxVec_20061116(cmplx(zsOld, 0.0d0, double)&
!      , zsNewDoubCmplx, xsNew, ysNew&
!      , xStartOld, yStartOld, xFinisOld, yFinisOld, style, nullElements)
    call interpolate2dDoubCmplxVec(cmplx(zsOld, 0.0d0, double)&
      , zsNewDoubCmplx, xsNew, ysNew&
      , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
      , nullElements)

    zsNew = real(zsNewDoubCmplx, single)

  end subroutine interpolate2dDoubleVec

  !---------------------------------------------------------------------
  subroutine interpolate2dSingCmplxVec(zsOld, zsNew, xsNew, ysNew&
  , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
  , nullElements)
  !---------------------------------------------------------------------

    implicit none

    complex(single), intent(in)            :: zsOld(:,:)
    complex(single), intent(out)           :: zsNew(:,:)
    real(single),    intent(in),  optional :: xsNew(size(zsNew,1)), ysNew(size(zsNew,2))
    real(single),    intent(in),  optional :: xStartOld, yStartOld,&
                                              xFinisOld, yFinisOld
    integer,         intent(in),  optional :: style
    logical,         intent(in),  optional :: forceNewStart
    logical,         intent(out), optional :: nullElements(size(zsNew,1),size(zsNew,2))

    complex(double), allocatable :: zsNewDoubCmplx(:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(zsNewDoubCmplx(size(zsNew,1),size(zsNew,2)))

    if (  present(xStartOld) .and. present(yStartOld)&
    .and. present(xFinisOld) .and. present(yFinisOld)) then
      if (present(xsNew) .and. present(ysNew)) then
!        call interpolate2dDoubCmplxVec_20061116(cmplx(zsOld, kind=double)&
!          , zsNewDoubCmplx, real(xsNew, double), real(ysNew, double)&
!          , real(xStartOld, double), real(yStartOld, double)&
!          , real(xFinisOld, double), real(yFinisOld, double), style, nullElements)
        call interpolate2dDoubCmplxVec(cmplx(zsOld, kind=double)&
          , zsNewDoubCmplx, real(xsNew, double), real(ysNew, double)&
          , real(xStartOld, double), real(yStartOld, double)&
          , real(xFinisOld, double), real(yFinisOld, double), style, forceNewStart&
          , nullElements)

      else if (.not. (present(xsNew) .or. present(ysNew))) then
!        call interpolate2dDoubCmplxVec_20061116(cmplx(zsOld, kind=double)&
!          , zsNewDoubCmplx&
!          , xStartOld=real(xStartOld, double)&
!          , yStartOld=real(yStartOld, double)&
!          , xFinisOld=real(xFinisOld, double)&
!          , yFinisOld=real(yFinisOld, double)&
!          , style=style, nullElements=nullElements)
        call interpolate2dDoubCmplxVec(cmplx(zsOld, kind=double)&
          , zsNewDoubCmplx&
          , xStartOld=real(xStartOld, double)&
          , yStartOld=real(yStartOld, double)&
          , xFinisOld=real(xFinisOld, double)&
          , yFinisOld=real(yFinisOld, double)&
          , style=style, forceNewStart=forceNewStart, nullElements=nullElements)

      else
call error('badArgumentCombination',"You must call interpolate2d with &
&either both or the optional arguments xsNew and ysNew, or neither.")
      end if

    else if (.not.(present(xStartOld) .or. present(yStartOld)&
              .or. present(xFinisOld) .or. present(yFinisOld))) then
      if (present(xsNew) .and. present(ysNew)) then
!        call interpolate2dDoubCmplxVec(_20061116cmplx(zsOld, kind=double)&
!          , zsNewDoubCmplx, real(xsNew, double), real(ysNew, double)&
!          , style=style, nullElements=nullElements)
        call interpolate2dDoubCmplxVec(cmplx(zsOld, kind=double)&
          , zsNewDoubCmplx, real(xsNew, double), real(ysNew, double)&
          , style=style, forceNewStart=forceNewStart, nullElements=nullElements)

      else if (.not. (present(xsNew) .or. present(ysNew))) then
!        call interpolate2dDoubCmplxVec_20061116(cmplx(zsOld, kind=double)&
!          , zsNewDoubCmplx, style=style, nullElements=nullElements)
        call interpolate2dDoubCmplxVec(cmplx(zsOld, kind=double)&
          , zsNewDoubCmplx, style=style, forceNewStart=forceNewStart&
          , nullElements=nullElements)

      else
call error('badArgumentCombination',"You must call interpolate2d with &
&either both or the optional arguments xsNew and ysNew, or neither.")
      end if

    else
call error('badArgumentCombination',"You must call interpolate2d with &
&either all the optional arguments xStartOld, yStartOld, xFinisOld and &
&yFinisOld, or none.")
    end if
    zsNew = cmplx(zsNewDoubCmplx, kind=double)
!AI NAG 5.2 porting
!kind=single changed to kind=double



  end subroutine interpolate2dSingCmplxVec

  !---------------------------------------------------------------------
  subroutine interpolate2dDoubCmplxVec(zsOld, zsNew, xsNew, ysNew&
  , xStartOld, yStartOld, xFinisOld, yFinisOld, style, forceNewStart&
  , nullElements)
  !---------------------------------------------------------------------

!***** NOTE that if you supply new values for any of zsOld, xStartOld, yStartOld, xFinisOld or yFinisOld, you MUST ALSO set forceNewStart=T.

    implicit none

    complex(double), intent(in)            :: zsOld(:,:)
    complex(double), intent(out)           :: zsNew(:,:)
    real(double),    intent(in),  optional :: xsNew(size(zsNew,1)), ysNew(size(zsNew,2))
    real(double),    intent(in),  optional :: xStartOld, yStartOld,&
                                              xFinisOld, yFinisOld
    integer,         intent(in),  optional :: style
    logical,         intent(in),  optional :: forceNewStart
    logical,         intent(out), optional :: nullElements(size(zsNew,1),size(zsNew,2))

    integer ::&
      xiSizeOld, xiSizeNew, yiSizeOld, yiSizeNew, xi, yi,&
      localStyle, xxi, yyi, i
!, lastBinNum
    integer, allocatable :: xBinNumbers(:), yBinNumbers(:)
    logical, allocatable :: xNulls(:), yNulls(:), localNulls(:,:)
    real(double), allocatable, save ::&
      localXsOld(:), localYsOld(:),&
      gradsReal(:,:,:), bigradsReal(:,:), gradsImag(:,:,:), bigradsImag(:,:)
    real(double), allocatable ::&
      localXsNew(:), localYsNew(:),&
!      localXsOld(:), localYsOld(:),&
!      gradsReal(:,:,:), bigradsReal(:,:), gradsImag(:,:,:), bigradsImag(:,:),&    
      xFracs(:), yFracs(:), xFracs2(:), xFracs3(:), yFracs2(:), yFracs3(:)
    real(double), save :: dxOld, dyOld, localXFinisOld, localYFinisOld, wt2(16,16)
    real(double) ::&
      localXStartOld, localYStartOld,&
!      dxOld, dyOld, localXFinisOld, localYFinisOld, wt2(16,16),&
      zNewReal, zNewImag, coeffs(4,4)
    logical :: localForceNewStart

    logical, save :: firstTime = .true. ! I know I don't really need the 'save' here but it makes it explicit.
    integer, save :: lastXBinNum, lastYBinNum
logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  interpolate2dDoubCmplxVec aaa'
if (testFlag) print *, '  firstTime:', firstTime

    localForceNewStart = decideOption(forceNewStart, .false.)
    localXStartOld = decideOption(xStartOld, 0.0d0)
    localYStartOld = decideOption(yStartOld, 0.0d0)

!    if (localForceNewStart) firstTime = .true.

    xiSizeOld = size(zsOld,1)
    xiSizeNew = size(zsNew,1)
    yiSizeOld = size(zsOld,2)
    yiSizeNew = size(zsNew,2)

    if (xiSizeOld < 2 .or. yiSizeOld < 2)&
call error('tooFewOldXOrYValues',"ssclib:")

    if (localForceNewStart .and. .not.firstTime)&
      deallocate(localXsOld, localYsOld)
    if (firstTime .or. localForceNewStart)&
      allocate(localXsOld(xiSizeOld),localYsOld(yiSizeOld))

    allocate(xBinNumbers(xiSizeNew), yBinNumbers(yiSizeNew))
    allocate(xNulls(xiSizeNew), yNulls(yiSizeNew), localNulls(xiSizeNew, yiSizeNew))
!    allocate(localXsOld(xiSizeOld),localYsOld(yiSizeOld))
    allocate(localXsNew(xiSizeNew),localYsNew(yiSizeNew))

    if (firstTime .or. localForceNewStart) then
      if (present(xFinisOld)) then
if (xFinisOld <= localXStartOld) call error('bug?',"")
        localXsOld = fracVal(localXStartOld, xFinisOld&
          , (/(i/real(xiSizeOld-1, double), i=0,xiSizeOld-1)/)) ! in ssclib/ssc_misc.
        dxOld = (xFinisOld - localXStartOld) / real(xiSizeOld-1, double)
      else
        localXsOld = localXStartOld + (/(real(i, double), i=0,xiSizeOld-1)/)
        dxOld = 1.0d0
      end if
      localXFinisOld = localXsOld(xiSizeOld)
    end if

    if (present(xsNew)) then
      localXsNew = xsNew
    else
      localXsNew = fracVal(localXStartOld, localXFinisOld&
        , (/(i/real(xiSizeNew-1, double), i=0,xiSizeNew-1)/)) ! in ssclib/ssc_misc.
    end if

if (testFlag) print *, 'size of xsOld and xsNew:', xiSizeOld, xiSizeNew
if (testFlag) print *, 'last vals xsOld and xsNew:', localXsOld(xiSizeOld), localXsNew(xiSizeNew)

    if (firstTime .or. localForceNewStart) lastXBinNum = 0
    do xi = 1, xiSizeNew
      lastXBinNum = getBinNumber(localXsOld, localXsNew(xi), lastXBinNum)
      xBinNumbers(xi) = lastXBinNum
    end do

if (testFlag) print *, 'last bin num:', xBinNumbers(xiSizeNew), '. No good?', (xBinNumbers(xiSizeNew)>=xiSizeOld)

    if (firstTime .or. localForceNewStart) then
      if (present(yFinisOld)) then
if (yFinisOld <= localYStartOld) call error('bug?',"")
        localYsOld = fracVal(localYStartOld, yFinisOld&
          , (/(i/real(yiSizeOld-1, double), i=0,yiSizeOld-1)/)) ! in ssclib/ssc_misc.
        dyOld = (yFinisOld - localYStartOld) / real(yiSizeOld-1, double)
      else
        localYsOld = localYStartOld + (/(real(i, double), i=0,yiSizeOld-1)/)
        dyOld = 1.0d0
      end if
      localYFinisOld = localYsOld(yiSizeOld)
    end if

    if (present(ysNew)) then
      localYsNew = ysNew
    else
      localYsNew = fracVal(localYStartOld, localYFinisOld&
        , (/(i/real(yiSizeNew-1, double), i=0,yiSizeNew-1)/)) ! in ssclib/ssc_misc.
    end if

    if (firstTime .or. localForceNewStart) lastYBinNum = 0
    do yi = 1, yiSizeNew
      lastYBinNum = getBinNumber(localYsOld, localYsNew(yi), lastYBinNum)
      yBinNumbers(yi) = lastYBinNum
    end do

    xNulls = .false. ! default
    where(xBinNumbers < 1 .or. xBinNumbers >= xiSizeOld)&
      xNulls = .true.

if (testFlag) print *, 'last x null?', xNulls(xiSizeNew)

    yNulls = .false. ! default
    where(yBinNumbers < 1 .or. yBinNumbers >= yiSizeOld)&
      yNulls = .true.

    localNulls = spread(xNulls, 2, yiSizeNew) .or. spread(yNulls, 1, xiSizeNew)

    allocate(xFracs(xiSizeNew))
    xFracs = 0.0 ! default
    do xi = 1, xiSizeNew
    if (xNulls(xi)) cycle
      xFracs(xi) = (localXsNew(xi) - localXsOld(xBinNumbers(xi))) / dxOld
    end do

    allocate(yFracs(yiSizeNew))
    yFracs = 0.0 ! default
    do yi = 1, yiSizeNew
    if (yNulls(yi)) cycle
      yFracs(yi) = (localYsNew(yi) - localYsOld(yBinNumbers(yi))) / dyOld
    end do

    zsNew = (0.0d0,0.0d0) ! default

    localStyle = decideOption(style, INTERP_LINEAR)
    select case (localStyle)
    case(INTERP_LINEAR)
      do xi = 1, xiSizeNew
!print *, xi, xFracs(xi), xBinNumbers(xi)
        do yi = 1, yiSizeNew
        if (localNulls(xi,yi)) cycle
          xxi = xBinNumbers(xi)
          yyi = yBinNumbers(yi)
!if (yi==90) then
!  tst=.true.
!else
!  tst=.false.
!end if

          zsNew(xi,yi) = bilinearInterpDbleCmplx(localXsOld(xxi:xxi+1)&
            , localYsOld(yyi:yyi+1), zsOld(xxi:xxi+1,yyi:yyi+1)&
            , localXsNew(xi), localYsNew(yi))!, tst)
        end do
      end do

    case(INTERP_SPLINE)
      if (localForceNewStart .and. .not.firstTime) then
        deallocate(gradsReal, bigradsReal, gradsImag, bigradsImag)
      end if

      if (firstTime .or. localForceNewStart) then
        allocate(gradsReal(xiSizeOld,yiSizeOld,2))
        allocate(bigradsReal(xiSizeOld,yiSizeOld))
        allocate(gradsImag(xiSizeOld,yiSizeOld,2))
        allocate(bigradsImag(xiSizeOld,yiSizeOld))
        call estimateGradsForBicubicDble(real(zsOld),  gradsReal, bigradsReal&
          , dxOld, dyOld)
        call estimateGradsForBicubicDble(aimag(zsOld), gradsImag, bigradsImag&
          , dxOld, dyOld)
        wt2 = loadWeightsForBicubic()
      end if

      allocate(xFracs2(xiSizeNew), xFracs3(xiSizeNew))
      xFracs2 = xFracs * xFracs
      xFracs3 = xFracs2 * xFracs
      allocate(yFracs2(yiSizeNew), yFracs3(yiSizeNew))
      yFracs2 = yFracs * yFracs
      yFracs3 = yFracs2 * yFracs

      do xi = 1, xiSizeNew
        do yi = 1, yiSizeNew
        if (localNulls(xi,yi)) cycle
          xxi = xBinNumbers(xi)
          yyi = yBinNumbers(yi)
          coeffs = calcCoeffsForBicubicDble(real(zsOld(xxi:xxi+1,yyi:yyi+1))&
            , gradsReal(xxi:xxi+1,yyi:yyi+1,:), bigradsReal(xxi:xxi+1,yyi:yyi+1)&
            , wt2)
          zNewReal = bicubicInterpDouble(xFracs3(xi), xFracs2(xi)&
            , xFracs(xi), yFracs3(yi), yFracs2(yi), yFracs(yi), coeffs)

          coeffs = calcCoeffsForBicubicDble(aimag(zsOld(xxi:xxi+1,yyi:yyi+1))&
            , gradsImag(xxi:xxi+1,yyi:yyi+1,:), bigradsImag(xxi:xxi+1,yyi:yyi+1)&
            , wt2)
          zNewImag = bicubicInterpDouble(xFracs3(xi), xFracs2(xi)&
            , xFracs(xi), yFracs3(yi), yFracs2(yi), yFracs(yi), coeffs)

          zsNew(xi,yi) = cmplx(zNewReal, zNewImag)
        end do
      end do

    case default
call error('badStyle',"The value of the style argument was not recognized.")
    end select

    if (present(nullElements))&
      nullElements = localNulls

    firstTime = .false.

  end subroutine interpolate2dDoubCmplxVec

  !---------------------------------------------------------------------
  function bilinearInterpSingle(xs, ys, zs, xIn, yIn) result(zOut)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: xs(2), ys(2), xIn, yIn
    real(single), intent(in) :: zs(2,2)
    real(single)             :: zOut

    complex(double) :: zOutDbleCmplx

    !-------------------------------------------------------------------
    ! Start

    zOutDbleCmplx = bilinearInterpDbleCmplx(real(xs, double)&
      , real(ys, double), cmplx(zs, 0.0, double), real(xIn, double)&
      , real(yIn, double)) 
    zOut = real(zOutDbleCmplx, single)

  end function bilinearInterpSingle

  !---------------------------------------------------------------------
  function bilinearInterpDouble(xs, ys, zs, xIn, yIn) result(zOut)
  !---------------------------------------------------------------------

    implicit none

    real(double),    intent(in) :: xs(2), ys(2), xIn, yIn
    real(double), intent(in) :: zs(2,2)
    real(double)             :: zOut

    complex(double) :: zOutDbleCmplx

    !-------------------------------------------------------------------
    ! Start

    zOutDbleCmplx = bilinearInterpDbleCmplx(real(xs, double)&
      , real(ys, double), cmplx(zs, 0.0, double), real(xIn, double)&
      , real(yIn, double)) 
    zOut = real(zOutDbleCmplx, kind=double)

  end function bilinearInterpDouble

  !---------------------------------------------------------------------
  function bilinearInterpSingCmplx(xs, ys, zs, xIn, yIn) result(zOut)
  !---------------------------------------------------------------------

    implicit none

    real(single),    intent(in) :: xs(2), ys(2), xIn, yIn
    complex(single), intent(in) :: zs(2,2)
    complex(single)             :: zOut

    complex(double) :: zOutDbleCmplx

    !-------------------------------------------------------------------
    ! Start

    zOutDbleCmplx = bilinearInterpDbleCmplx(real(xs, double)&
      , real(ys, double), cmplx(zs, kind=double), real(xIn, double)&
      , real(yIn, double)) 
    zOut = cmplx(zOutDbleCmplx, kind=double)
!AI NAG 5.2 porting
!kind=single changed to kind=double


  end function bilinearInterpSingCmplx

  !---------------------------------------------------------------------
  function bilinearInterpDbleCmplx(xs, ys, zs, xIn, yIn) result(zOut)
  !---------------------------------------------------------------------

    implicit none

    real(double),    intent(in) :: xs(2), ys(2), xIn, yIn
    complex(double), intent(in) :: zs(2,2)
!logical, intent(in), optional :: tst
    complex(double)             :: zOut

    real(double) :: ax, bx, xDenom, ay, by, yDenom

    !-------------------------------------------------------------------
    ! Start

    xDenom = xs(2) - xs(1)
    yDenom = ys(2) - ys(1)
    ax = (xs(2) - xIn)  / xDenom
    ay = (ys(2) - yIn)  / yDenom
!    bx = (xIn   - xs(1)) / xDenom
!    by = (yIn   - ys(1)) / yDenom
!*** unnecessary division operations.
bx = 1.0d0 - ax
by = 1.0d0 - ay

    zOut = ax * ay * zs(1, 1)&
      +    bx * ay * zs(2, 1)&
      +    bx * by * zs(2, 2)&
      +    ax * by * zs(1, 2)

!if (present(tst)) then
!  if (tst) then
!    print *, bx
!  end if
!end if

  end function bilinearInterpDbleCmplx

  !---------------------------------------------------------------------
  subroutine estimateGradsForBicubicSing(zs, grads, bigrads, dx, dy)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: zs(:,:)
    real(single), intent(out)          :: grads(  size(zs,1),size(zs,2),2),&
                                          bigrads(size(zs,1),size(zs,2))
    real(single), intent(in), optional :: dx, dy

    integer :: nx, ny
    real(double), allocatable :: gradsDble(:,:,:), bigradsDble(:,:)

    !-------------------------------------------------------------------
    ! Start

    nx = size(zs,1)
    ny = size(zs,2)
    allocate(gradsDble(nx, ny, 2), bigradsDble(nx, ny))

    call estimateGradsForBicubicDble(real(zs, double), gradsDble, bigradsDble&
      , real(dx, double), real(dy, double))

    grads = real(gradsDble, single)
    bigrads = real(bigradsDble, single)

  end subroutine estimateGradsForBicubicSing

  !---------------------------------------------------------------------
  subroutine estimateGradsForBicubicDble(zs, grads, bigrads, dx, dy)
  !---------------------------------------------------------------------
  ! Performs a finite-difference estimate of (dz/dx,dz/dy) and dz2/dxdy.
  ! The former is returned in (grads(:,:,1),grads(:,:,2)) and the latter
  ! in bigrads. Square subarrays of grads and bigrads are then suitable
  ! for input to hermiteSpline2dCalcCoeffs.
  !
  ! Technically speaking the formulae for the non-edge values were taken
  ! from Press et al, but this is freshman numerical analysis.

    implicit none

    real(double), intent(in)           :: zs(:,:)
    real(double), intent(out)          :: grads(  size(zs,1),size(zs,2),2),&
                                          bigrads(size(zs,1),size(zs,2))
    real(double), intent(in), optional :: dx, dy

    integer :: nx, ny
    real(double) :: localDx, localDy

    !-------------------------------------------------------------------
    ! Start

    nx = size(zs,1)
    ny = size(zs,2)

    localDx = decideOption(dx, 1.0d0) ! in ssc_misc.
    localDy = decideOption(dy, 1.0d0) ! in ssc_misc.

    ! dz/dx:
    !
    grads(1,     :,1) = (zs(2,   :) - zs(1,     :)) / localDx
    grads(2:nx-1,:,1) = (zs(3:nx,:) - zs(1:nx-2,:)) / localDx / 2.0d0
    grads(  nx,  :,1) = (zs(  nx,:) - zs(  nx-1,:)) / localDx

    ! dz/dy:
    !
    grads(:,1,     2) = (zs(:,2   ) - zs(:,1     )) / localDy
    grads(:,2:ny-1,2) = (zs(:,3:ny) - zs(:,1:ny-2)) / localDy / 2.0d0
    grads(:,  ny,  2) = (zs(:,  ny) - zs(:,  ny-1)) / localDy

    ! dz2/dxdy:
    !
    bigrads(1,          1) = (zs(2,   2     ) + zs(1,     1     )&
                            - zs(2,   1     ) - zs(1,     2     ))&
      / localDx / localDy
    bigrads(2:nx-1,     1) = (zs(3:nx,2     ) + zs(1:nx-2,1     )&
                            - zs(3:nx,1     ) - zs(1:nx-2,2     ))&
      / localDx / localDy / 2.0d0
    bigrads(  nx,       1) = (zs(  nx,2     ) + zs(  nx-1,1     )&
                            - zs(  nx,1     ) - zs(  nx-1,2     ))&
      / localDx / localDy
    bigrads(1,     2:ny-1) = (zs(2,   3:ny  ) + zs(1,     1:ny-2)&
                            - zs(2,   1:ny-2) - zs(1,     3:ny  ))&
      / localDx / localDy / 2.0d0
    bigrads(2:nx-1,2:ny-1) = (zs(3:nx,3:ny  ) + zs(1:nx-2,1:ny-2)&
                            - zs(3:nx,1:ny-2) - zs(1:nx-2,3:ny  ))&
      / localDx / localDy / 4.0d0
    bigrads(  nx,  2:ny-1) = (zs(  nx,3:ny  ) + zs(  nx-1,1:ny-2)&
                            - zs(  nx,1:ny-2) - zs(  nx-1,3:ny  ))&
      / localDx / localDy / 2.0d0
    bigrads(1,       ny  ) = (zs(2,     ny  ) + zs(1,       ny-1)&
                            - zs(2,     ny-1) - zs(1,       ny  ))&
      / localDx / localDy
    bigrads(2:nx-1,  ny  ) = (zs(3:nx,  ny  ) + zs(1:nx-2,  ny-1)&
                            - zs(3:nx,  ny-1) - zs(1:nx-2,  ny  ))&
      / localDx / localDy / 2.0d0
    bigrads(  nx,    ny  ) = (zs(  nx,  ny  ) + zs(  nx-1,  ny-1)&
                            - zs(  nx,  ny-1) - zs(  nx-1,  ny  ))&
      / localDx / localDy

  end subroutine estimateGradsForBicubicDble

  !---------------------------------------------------------------------
  function loadWeightsForBicubic() result(wt2)
  !---------------------------------------------------------------------
  ! From Numerical Recipes (Press et al).

    implicit none

    real(double) :: wt2(16,16)

    !-------------------------------------------------------------------
    ! Start

    wt2(:, 1) = (/ 1, 0,-3, 2, 0, 0, 0, 0,-3, 0, 9,-6, 2, 0,-6, 4/)
    wt2(:, 2) = (/ 0, 0, 0, 0, 0, 0, 0, 0, 3, 0,-9, 6,-2, 0, 6,-4/)
    wt2(:, 3) = (/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9,-6, 0, 0,-6, 4/)
    wt2(:, 4) = (/ 0, 0, 3,-2, 0, 0, 0, 0, 0, 0,-9, 6, 0, 0, 6,-4/)

    wt2(:, 5) = (/ 0, 0, 0, 0, 1, 0,-3, 2,-2, 0, 6,-4, 1, 0,-3, 2/)
    wt2(:, 6) = (/ 0, 0, 0, 0, 0, 0, 0, 0,-1, 0, 3,-2, 1, 0,-3, 2/)
    wt2(:, 7) = (/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,-3, 2, 0, 0, 3,-2/)
    wt2(:, 8) = (/ 0, 0, 0, 0, 0, 0, 3,-2, 0, 0,-6, 4, 0, 0, 3,-2/)
    wt2(:, 9) = (/ 0, 1,-2, 1, 0, 0, 0, 0, 0,-3, 6,-3, 0, 2,-4, 2/)
    wt2(:,10) = (/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 3,-6, 3, 0,-2, 4,-2/)
    wt2(:,11) = (/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,-3, 3, 0, 0, 2,-2/)
    wt2(:,12) = (/ 0, 0,-1, 1, 0, 0, 0, 0, 0, 0, 3,-3, 0, 0,-2, 2/)

    wt2(:,13) = (/ 0, 0, 0, 0, 0, 1,-2, 1, 0,-2, 4,-2, 0, 1,-2, 1/)
    wt2(:,14) = (/ 0, 0, 0, 0, 0, 0, 0, 0, 0,-1, 2,-1, 0, 1,-2, 1/)
    wt2(:,15) = (/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,-1, 0, 0,-1, 1/)
    wt2(:,16) = (/ 0, 0, 0, 0, 0, 0,-1, 1, 0, 0, 2,-2, 0, 0,-1, 1/)

  end function loadWeightsForBicubic

  !---------------------------------------------------------------------
  function calcCoeffsForBicubicSing(y, grads, bigrads, wt2) result(coeffs)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: y(2,2), grads(2,2,2), bigrads(2,2)
    real(double), intent(in) :: wt2(16,16)
    real(single)             :: coeffs(4,4)

    real(double) :: coeffsDble(4,4)

    !-------------------------------------------------------------------
    ! Start

    coeffsDble = calcCoeffsForBicubicDble(real(y, double), real(grads, double)&
      , real(bigrads, double), wt2)
    coeffs = real(coeffsDble, single)

  end function calcCoeffsForBicubicSing

  !---------------------------------------------------------------------
  function calcCoeffsForBicubicDble(y, grads, bigrads, wt2) result(coeffs)
  !---------------------------------------------------------------------
  ! From Numerical Recipes (Press et al).

    implicit none

    real(double), intent(in) :: y(2,2), grads(2,2,2), bigrads(2,2), wt2(16,16)
    real(double)             :: coeffs(4,4)

    real(double) :: x(16), k(16), xx
    integer :: i, j, n

    !-------------------------------------------------------------------
    ! Start

    x( 1) = y(1,1)
    x( 2) = y(2,1)
    x( 3) = y(2,2)
    x( 4) = y(1,2)
    x( 5) = grads(1,1,1)
    x( 6) = grads(2,1,1)
    x( 7) = grads(2,2,1)
    x( 8) = grads(1,2,1)
    x( 9) = grads(1,1,2)
    x(10) = grads(2,1,2)
    x(11) = grads(2,2,2)
    x(12) = grads(1,2,2)
    x(13) = bigrads(1,1)
    x(14) = bigrads(2,1)
    x(15) = bigrads(2,2)
    x(16) = bigrads(1,2)

    do i = 1, 16
      xx = 0.0
      do j = 1, 16
        xx = xx + wt2(i,j) * x(j)
      end do
      k(i) = xx
    end do

    n = 0
    do i = 1, 4
      do j = 1, 4
        n = n + 1
        coeffs(i,j) = k(n)
      end do
    end do

  end function calcCoeffsForBicubicDble

  !---------------------------------------------------------------------
  function bicubicInterpSingle(x3, x2, x, y3, y2, y, k) result(z)
  !---------------------------------------------------------------------
  ! From Numerical Recipes (Press et al).

    implicit none

    real(single), intent(in) :: x3, x2, x, y3, y2, y, k(4,4)
    real(single)             :: z

    !-------------------------------------------------------------------
    ! Start

    z = k(1,1)    + k(1,2)*y    + k(1,3)*y2    + k(1,4)*y3&
      + k(2,1)*x  + k(2,2)*x*y  + k(2,3)*x*y2  + k(2,4)*x*y3&
      + k(3,1)*x2 + k(3,2)*x2*y + k(3,3)*x2*y2 + k(3,4)*x2*y3&
      + k(4,1)*x3 + k(4,2)*x3*y + k(4,3)*x3*y2 + k(4,4)*x3*y3

   end function bicubicInterpSingle

  !---------------------------------------------------------------------
  function bicubicInterpDouble(x3, x2, x, y3, y2, y, k) result(z)
  !---------------------------------------------------------------------
  ! From Numerical Recipes (Press et al).

    implicit none

    real(double), intent(in) :: x3, x2, x, y3, y2, y, k(4,4)
    real(double)             :: z

    !-------------------------------------------------------------------
    ! Start

    z = k(1,1)    + k(1,2)*y    + k(1,3)*y2    + k(1,4)*y3&
      + k(2,1)*x  + k(2,2)*x*y  + k(2,3)*x*y2  + k(2,4)*x*y3&
      + k(3,1)*x2 + k(3,2)*x2*y + k(3,3)*x2*y2 + k(3,4)*x2*y3&
      + k(4,1)*x3 + k(4,2)*x3*y + k(4,3)*x3*y2 + k(4,4)*x3*y3

   end function bicubicInterpDouble


  !---------------------------------------------------------------------
  subroutine interpolateInPolarCoord(thetaSamplesVector, phiSamplesVector&
  , zSamplesArray, thetaMap, phiMap, interpolatedZMap)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) ::&
      thetaSamplesVector(:), phiSamplesVector(:),&
      zSamplesArray(size(thetaSamplesVector),size(phiSamplesVector))
    real(double), intent(in) ::&
      thetaMap(:,:), phiMap(size(thetaMap, 1), size(thetaMap, 2))
    real(single), intent(out) ::&
      interpolatedZMap(size(thetaMap, 1), size(thetaMap, 2))

    integer(int16) ::&
      numThetaSamples, numPhiSamples,&
      xsize, ysize, loPhiI, hiPhiI,&
      xi, yi, phiI, i, j
    real(single), allocatable ::&
      secondDerivWrtTheta(:,:)
    real(single) :: thetaArcsec, phi, loPhiZValue, hiPhiZValue, phiFrac

    !-------------------------------------------------------------------
    ! Start

    numThetaSamples = size(thetaSamplesVector)
    numPhiSamples   = size(phiSamplesVector)
    ! NOTE these all must be >= 2.

    xsize = size(thetaMap, 1)
    ysize = size(thetaMap, 2)

    if  (minval(thetaMap) < thetaSamplesVector(1)) then
print *, 'minval(thetaMap), thetaSamplesVector(1): ', minval(thetaMap), thetaSamplesVector(1)
call error('bug','')
    end if

    if  (minval(phiMap) < phiSamplesVector(1)&
    .or. maxval(phiMap) > phiSamplesVector(numPhiSamples)) then
print *, 'minval(phiMap), phiSamplesVector(1): ', minval(phiMap), phiSamplesVector(1)
print *, 'maxval(phiMap), phiSamplesVector(n): ', maxval(phiMap), phiSamplesVector(numThetaSamples)
call error('bug','')
    end if

    loPhiI =  largestTrueIndex(logical(phiSamplesVector <= minval(phiMap),bool))
    hiPhiI = smallestTrueIndex(logical(phiSamplesVector >= maxval(phiMap),bool))
    ! These are in ssclib/array_utils.

    if (loPhiI == hiPhiI) then ! unlikely as it implies that
      ! there is only 1 value of phi among the image pixels, and
      ! also that that value lies exactly on one of the samples
      ! in phiSamplesVector. Nevertheless, could happen I suppose.

      if (hiPhiI < numPhiSamples) then
        hiPhiI = hiPhiI + 1
      else
        loPhiI = loPhiI - 1
      end if
    end if

    ! Preload the 2nd derivatives wrt theta for the cubic spline:

    allocate(secondDerivWrtTheta(numThetaSamples, loPhiI:hiPhiI))
    do phiI = loPhiI, hiPhiI
!      call calcCubicSplines(thetaSamplesVector, zSamplesArray(:, phiI)&
!        , loBoundType='fixed1st', loBoundValue=0.0&
!        , hiBoundType='fixed2nd', hiBoundValue=0.0& ! only an approx,
!                                                    ! don't know the
!                                                    ! correct outer BC.
!        , y2=secondDerivWrtTheta(:, phiI)&
!      )
      call calcY2ForCubicSplinesSing(thetaSamplesVector, zSamplesArray(:, phiI)&
        , secondDerivWrtTheta(:, phiI)&
        , loBoundType=SPLINE_FIXED_1ST, loBoundValue=0.0&
        , hiBoundType=SPLINE_FIXED_2ND, hiBoundValue=0.0& ! only an approx,
                                                    ! don't know the
                                                    ! correct outer BC.
      )
    end do

    ! Now interpolate the z samples array for all image pixels:

    interpolatedZMap = 0.0
    do xi = 1, xsize
      do yi = 1, ysize
        thetaArcsec = thetaMap(xi, yi)	! type conversion from
        phi = phiMap(xi, yi)		! double to single.
      if (thetaArcsec > thetaSamplesVector(numThetaSamples)) cycle

        ! Get the indices of the lower LH corner of the theta-phi
        ! rectangle in theta-phi space that contains both thetaArcsec
        ! and phi:

        i = largestTrueIndex(logical(thetaSamplesVector <= thetaArcsec, bool))
        if (i == numThetaSamples) i = numThetaSamples - 1
        j = largestTrueIndex(logical(phiSamplesVector <= phi, bool))
        if (j == hiPhiI) j = hiPhiI - 1

        ! Get the cubic-spline interpolation at each of the two
        ! bracketting phi values:

!        call cubicSplineInterpolate(thetaSamplesVector(i:i+1)&
!          , zSamplesArray(i:i+1, j)&
!          , secondDerivWrtTheta(i:i+1, j),   thetaArcsec, loPhiZValue)
!        call cubicSplineInterpolate(thetaSamplesVector(i:i+1)&
!          , zSamplesArray(i:i+1, j+1)&
!          , secondDerivWrtTheta(i:i+1, j+1), thetaArcsec, hiPhiZValue)
loPhiZValue = cubicSplineInterpSingle(thetaSamplesVector(i:i+1)&
          , zSamplesArray(i:i+1, j)&
          , secondDerivWrtTheta(i:i+1, j),   thetaArcsec)
hiPhiZValue = cubicSplineInterpSingle(thetaSamplesVector(i:i+1)&
          , zSamplesArray(i:i+1, j+1)&
          , secondDerivWrtTheta(i:i+1, j+1), thetaArcsec)

        ! Finally do a linear interpolation between these two phi values:

        phiFrac = (phi - phiSamplesVector(j))&
          / (phiSamplesVector(j+1) - phiSamplesVector(j))
        interpolatedZMap(xi, yi) = fracVal(loPhiZValue, hiPhiZValue, phiFrac)
      end do
    end do

    deallocate(secondDerivWrtTheta)

    if (any(thetaMap > thetaSamplesVector(numThetaSamples)))&
      call warning('thetaOutsideRange','')

  end subroutine interpolateInPolarCoord



!***** what to do with the rest of this stuff? Is it obsolete or should we (must we) resurrect some??




  !---------------------------------------------------------------------
  real function bilinearSample(zGridSamples, xSamples, ySamples, x, y&
  , pixelIsNull, returnedNullValue)
  !---------------------------------------------------------------------

    implicit none

    real(single),  intent(in) ::&
      xSamples(:),&
      ySamples(:),&
      zGridSamples(size(xSamples), size(ySamples)),&
      x,&
      y
    logical(bool), intent(in), optional :: pixelIsNull(size(xSamples)-1&
                                                     , size(ySamples)-1)
    real(single),  intent(in), optional :: returnedNullValue

    real(single) :: fracX, fracY
    integer :: xi, yi, xsize, ysize

    !-------------------------------------------------------------------
    ! Start

    ! Load default value:
    if (present(returnedNullValue)) then
      bilinearSample = returnedNullValue
    else
      bilinearSample = 0.0
    end if

    xsize = size(xSamples)
    ysize = size(ySamples)
  if (xsize <= 1 .or. ysize <= 1) return

    if ( x < xSamples(1) .or. x > xSamples(xsize)&
    .or. y < ySamples(1) .or. y > ySamples(ysize))&
  return

    xi = getBinNumber(xSamples, x)
    yi = getBinNumber(ySamples, y)

    if (present(pixelIsNull)) then
  if (pixelIsNull(xi, yi)) return
    end if

    fracX = (x - xSamples(xi)) / (xSamples(xi+1) - xSamples(xi))
    fracY = (y - ySamples(yi)) / (ySamples(yi+1) - ySamples(yi))
    bilinearSample = bilinearValueSingle(zGridSamples(xi:xi+1, yi:yi+1)&
      , fracX, fracY)

  end function bilinearSample

  !---------------------------------------------------------------------
  real(single) function bilinearValueSingle(zAtCorners, fracX, fracY)
  !---------------------------------------------------------------------

!**** NOTE the z indices are the right way around here!! Not elsewhere.

    implicit none

    real(single), intent(in) :: zAtCorners(2,2), fracX, fracY

    !-------------------------------------------------------------------
    ! Start

    bilinearValueSingle = (1.0 - fracX) * (1.0 - fracY) * zAtCorners(1, 1)&
             +             fracX  * (1.0 - fracY) * zAtCorners(2, 1)&
             +             fracX         * fracY  * zAtCorners(2, 2)&
             +      (1.0 - fracX)        * fracY  * zAtCorners(1, 2)

  end function bilinearValueSingle

  !---------------------------------------------------------------------
  real(double) function bilinearValueDouble(zAtCorners, fracX, fracY)
  !---------------------------------------------------------------------

!**** NOTE the z indices are the right way around here!! Not elsewhere.

    implicit none

    real(double), intent(in) :: zAtCorners(2,2), fracX, fracY

    !-------------------------------------------------------------------
    ! Start

    bilinearValueDouble = (1.0d0 - fracX) * (1.0d0 - fracY) * zAtCorners(1, 1)&
             +               fracX  * (1.0d0 - fracY) * zAtCorners(2, 1)&
             +               fracX           * fracY  * zAtCorners(2, 2)&
             +      (1.0d0 - fracX)          * fracY  * zAtCorners(1, 2)

  end function bilinearValueDouble

  !---------------------------------------------------------------------
  complex(single) function bilinearValueSingComplex(zAtCorners, fracX, fracY)
  !---------------------------------------------------------------------

!**** NOTE the z indices are the right way around here!! Not elsewhere.

    implicit none

    complex(single), intent(in) :: zAtCorners(2,2)
    real(single),    intent(in) :: fracX, fracY

    !-------------------------------------------------------------------
    ! Start

    bilinearValueSingComplex = (1.0 - fracX) * (1.0 - fracY) * zAtCorners(1, 1)&
             +             fracX  * (1.0 - fracY) * zAtCorners(2, 1)&
             +             fracX         * fracY  * zAtCorners(2, 2)&
             +      (1.0 - fracX)        * fracY  * zAtCorners(1, 2)

  end function bilinearValueSingComplex

  !---------------------------------------------------------------------
  complex(double) function bilinearValueDoubComplex(zAtCorners, fracX, fracY)
  !---------------------------------------------------------------------

!**** NOTE the z indices are the right way around here!! Not elsewhere.

    implicit none

    complex(double), intent(in) :: zAtCorners(2,2)
    real(double),    intent(in) :: fracX, fracY

    !-------------------------------------------------------------------
    ! Start

    bilinearValueDoubComplex = (1.0d0 - fracX) * (1.0d0 - fracY) * zAtCorners(1, 1)&
             +               fracX  * (1.0d0 - fracY) * zAtCorners(2, 1)&
             +               fracX           * fracY  * zAtCorners(2, 2)&
             +      (1.0d0 - fracX)          * fracY  * zAtCorners(1, 2)

  end function bilinearValueDoubComplex

  !---------------------------------------------------------------------
  real function bilinearSampleVariance(zVarGrid, xVector, yVector, x, y)
  !---------------------------------------------------------------------
!***** these 2 routines were taken from package non_sas/srcori. I am not sure what they mean - I'd have to look at srcori to find out, and at the moment I don't want to take the time. I just store them here as non-public routines to conserve th code for a future time when it may be found of value. IMS 2006-10-16.

    implicit none

    real, intent(in) ::&
      zVarGrid(:,:),&
      xVector(size(zVarGrid, 1)),&
      yVector(size(zVarGrid, 2)),&
      x,&
      y

    real :: fracX, fracY
    integer :: xi, xxi, yi, yyi, xsize, ysize

    !-------------------------------------------------------------------
    ! Start

    bilinearSampleVariance = 0.0 ! default
    xsize = size(zVarGrid, 1)
    ysize = size(zVarGrid, 2)
    if (x .lt. xVector(1) .or. x .ge. xVector(xsize)&
    .or. y .lt. yVector(1) .or. y .ge. yVector(ysize))&
  return

    do xi = 2, xsize
      if (x .lt. xVector(xi)) then
        xxi = xi - 1
    exit
      end if
    end do
    do yi = 2, ysize
      if (y .lt. yVector(yi)) then
        yyi = yi - 1
    exit
      end if
    end do

    fracX = (x - xVector(xxi)) / (xVector(xxi+1) - xVector(xxi))
    fracY = (y - yVector(yyi)) / (yVector(yyi+1) - yVector(yyi))
    bilinearSampleVariance = bilinearValueVariance(zVarGrid(xxi:xxi+1, yyi:yyi+1)&
    , fracX, fracY)

  end function bilinearSampleVariance

  !---------------------------------------------------------------------
  real function bilinearValueVariance(zVar, fracX, fracY)
  !---------------------------------------------------------------------
!***** these 2 routines were taken from package non_sas/srcori. I am not sure what they mean - I'd have to look at srcori to find out, and at the moment I don't want to take the time. I just store them here as non-public routines to conserve th code for a future time when it may be found of value. IMS 2006-10-16.

    implicit none

    real, intent(in) :: zVar(2,2), fracX, fracY

    real :: carfX, carfY

    !-------------------------------------------------------------------
    ! Start

    carfX = 1.0 - fracX
    carfY = 1.0 - fracY
    bilinearValueVariance = (carfX * carfY) * (carfX * carfY) * zVar(1, 1)&
                     + (fracX * carfY) * (fracX * carfY) * zVar(1, 2)&
                     + (fracX * fracY) * (fracX * fracY) * zVar(2, 2)&
                     + (carfX * fracY) * (carfX * fracY) * zVar(2, 1)

  end function bilinearValueVariance

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module interpolations
