! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: source_cutouts
!
! Authors:  Ian Stewart - SSC/LUX
!        :  Masaaki Sakano - SSC/LUX
!
! Description: Returns semi-axes of an ellipse round each of a list of sources.
! 
! Comments: TODO:
! 
! $Id: source_cutouts.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module source_cutouts

  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use errorHandling

  use coordinate
  use psf_ims
  use wcs_aux
  use constants
  use ellipse

  implicit none 

  private 

  public initializeSrcDataT
  public getSourceCutouts
  public dumpSrcDataT

  interface initializeSrcDataT
    module procedure initializeSrcDataTScalar
    module procedure initializeSrcDataTVector
  end interface

  interface getSourceCutouts
    module procedure getSourceCutoutsFewArgs, getSourceCutoutsManyArgs
  end interface

  interface dumpSrcDataT
    module procedure dumpSrcDataTScalar
  end interface

  real, parameter :: ThLowerBkgFluxPerPixel = 0.0001	! BG_MAP set by
  ! emldetect (more strictly, esplinemap) has occasionally
  ! values of order 10^-5, whereas it should be much largedr.
  ! In those cases, an exceptionally large source area is produced
  ! by the task region, which calls this library.
  ! Those cases are apparently wrong.  Therefore this parameter gives
  ! a work-around for it; i.e., if the BG_MAP is the same as or smaller
  ! than this threshold, then it is treated like a zero BG_MAP.
  ! See SOC-SPR-6594.
  !
  !
  ! FYI, here is the excerpt from Georg Lamer's email:
  !
  !! I suspect that a bad esplinemap fit is the underlying reason for the
  !! problem. The example Steve sent, is a source within the core of M31.
  !! Probably the spline fit went wrong, when most of the pixels in the crowded
  !! field were masked out. 1E-5 is the minimum value the background
  !! map can have. I guess for the moment the region task needs to avoid
  !! this problem with a reasonable upper limit for the radius.
  !! However we also should look into improvements to esplinemap or its parameter
  !! seetings, I just checked and found that ~10000 sources in the 2XMM catalogue
  !! have a PN 0.5-1 keV background of 1E-5.
  !! 


  type, public :: SrcDataT
    integer(int16) ::&
      idNum
    real(single) ::&
!      xDet,&
!      yDet,&
      flux,&		! These three should generally be in compatible
      bkgFluxPerPixel,&	! units. Thus, if flux is cts/s, bkgFlux should
!      bkgFluxPerArcsec,&! be cts/s/pixel.
      xPixel,&
      yPixel
    character(10) ::&
      fluxUnit	! eg 'counts', 'rate', 'cgs'
    type(EllipseT) ::&	! Defined in ssclib/ellipse
      region
    real(double) ::&
xDet,&
yDet,&
      raDeg,&
      decDeg,&
      EeV,&
      thetaArcsec,&
      phi,&
      xTan,&
      yTan
    logical(bool) ::&
      isValid
  end type SrcDataT

contains 

  !---------------------------------------------------------------------
  subroutine initializeSrcDataTScalar(srcData)
  !---------------------------------------------------------------------

    implicit none

    type(SrcDataT), intent(out) :: srcData

    type(SrcDataT) :: srcDataVector(1)

    !-------------------------------------------------------------------
    ! Start

    call initializeSrcDataTVector(srcDataVector)
    srcData = srcDataVector(1)

  end subroutine initializeSrcDataTScalar

  !---------------------------------------------------------------------
  subroutine initializeSrcDataTVector(srcData)
  !---------------------------------------------------------------------

    implicit none

    type(SrcDataT), intent(out) :: srcData(:)

    !-------------------------------------------------------------------
    ! Start

    srcData%flux = 0.0
    srcData%bkgFluxPerPixel = 0.0
    srcData%xPixel = 0.0
    srcData%yPixel = 0.0
    srcData%xDet = 0.0
    srcData%yDet = 0.0
    srcData%fluxUnit = 'none'
    call initializeEllipseT(srcData%region)
    srcData%raDeg = 0.0d0
    srcData%decDeg = 0.0d0
    srcData%EeV = 0.0d0
    srcData%thetaArcsec = 0.0d0
    srcData%phi = 0.0d0
    srcData%xTan = 0.0d0
    srcData%yTan = 0.0d0
    srcData%isValid = .false.

  end subroutine initializeSrcDataTVector

  !---------------------------------------------------------------------
  subroutine getSourceCutoutsFewArgs(imgWcs, srcWcs, scAttitude&
  , expStartTimeStamp, fracBackground, srcData)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up with the correct instrument beforehand.
  !
  ! DESCRIPTION:
  !  This routine fills the various parameters in srcData.
  !  In doing that, this routine gives the PSF shape for the correct size,
  ! depending on its given background level, for each source
  !
  ! NOTE:
  !  srcData%raDeg and %decDeg in the argument are assumed to be already set.
  !  After this subroutine,
  !   srcData%xPixel, %yPixel, %xDet, %yDet, %thetaArcsec, %phi, %xTan, %yTan, %region
  !  are set (or overwritten, even if preset), unless it returns
  !  due to an erroneous situation.
  !  srcData%bkgFluxPerPixel may be overwritten if the input value
  !  is unrealistically small.

    implicit none

    type(WcsT),                   intent(in)    :: imgWcs,&
                                                   srcWcs
    real(double),                 intent(in)    :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(in)    :: scAttitude
    real(single),                 intent(in)    :: fracBackground
    type(SrcDataT),               intent(inout) :: srcData(:)

    integer :: numSources, srcNum
    ! logical(bool), allocatable :: mask(:)
    logical, allocatable :: mask(:)
    real(single), allocatable ::&
      bkgFluxesPerSquareArcsec(:),&
      ellipseLongAxisArcsec(:),&
      ellipseShrtAxisArcsec(:),&
      ellipseLongAxisTan(:),&
      ellipseShrtAxisTan(:),&
      ellipseRotAngle(:),&
      ellipseXampTan(:), ellipseYampTan(:),&
      peakFluxPerSquareArcsec(:)
    real(single) ::&
      integral,&
      averageBkgFluxPerPixel, srcPixelAreaArcsec
    real(double) ::&
      ctrFovXtan, ctrFovYtan,&
pixelSizeXWorld, pixelSizeYWorld

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcData)
    allocate(mask(numSources))
    allocate(bkgFluxesPerSquareArcsec(numSources))
    allocate(ellipseLongAxisArcsec(numSources))
    allocate(ellipseShrtAxisArcsec(numSources))
    allocate(ellipseLongAxisTan(numSources))
    allocate(ellipseShrtAxisTan(numSources))
    allocate(ellipseRotAngle(numSources))
    allocate(ellipseXampTan(numSources), ellipseYampTan(numSources))
    allocate(peakFluxPerSquareArcsec(numSources))

    !...................................................................
    ! Do some checks:

    if (all(srcData%flux <= 0.0)) then
      call warning('allSrcNonPositive', "ssclib:getSourceCutoutsFewArgs: None &
        &of the sources have rate values greater than zero.")
  return
    end if

    if (all(srcData%flux <= 0.0 .or. srcData%bkgFluxPerPixel <= 0.0)) then
      call warning('allSrcBkgNonPositive', "ssclib:getSourceCutoutsFewArgs: None &
        &of the positive-rate sources have a positive-valued local background.")
  return
    end if

    if (any(srcData%flux > 0.0 .and. srcData%bkgFluxPerPixel <= ThLowerBkgFluxPerPixel)) then
      call warning('nonPositiveSrcBkg', "ssclib:getSourceCutoutsFewArgs: &
        &Some of the background values of the source list are less than or &
        &equal to or unreasonably close to zero. These have been set to &
        &the average of the positive values.")

      mask = (srcData%flux > 0.0 .and. srcData%bkgFluxPerPixel > ThLowerBkgFluxPerPixel)
      averageBkgFluxPerPixel = sum(srcData%bkgFluxPerPixel, mask)&
        / real(count(mask))

      ! where(srcData%flux > 0.0 .and. srcData%bkgFluxPerPixel <= ThLowerBkgFluxPerPixel)&
      mask = .false.
      mask = (srcData%flux > 0.0 .and. srcData%bkgFluxPerPixel <= ThLowerBkgFluxPerPixel)

      if (verbosity(AppMsg) >= Verbose) then
        write(errStr, '("For the followings, the BGD value is set to be ", ES15.8)') &
           averageBkgFluxPerPixel
        call message(AppMsg, Verbose)
        do srcNum = 1, numSources
          if (mask(srcNum)) then
            if (srcData(srcNum)%xPixel == 0 .and. srcData(srcNum)%yPixel == 0) then
              write(errStr, '(" idNum=",i6,", bkgFluxPerPixel=",ES14.7,' &
                 //'", (raDeg,decDeg)=(",F12.8,", ",F12.8,")")') &
                   srcData(srcNum)%idNum, srcData(srcNum)%bkgFluxPerPixel &
                 , srcData(srcNum)%raDeg, srcData(srcNum)%decDeg
            else
              write(errStr, '(" idNum=",i6,", bkgFluxPerPixel=",ES14.7,' &
                 //'", (xPixel,yPixel)=(",F12.6,", ",F12.6,")")') &
                   srcData(srcNum)%idNum, srcData(srcNum)%bkgFluxPerPixel &
                 , srcData(srcNum)%xPixel, srcData(srcNum)%yPixel
            end if
            call message(AppMsg, Verbose)
! call dumpSrcDataT(srcData(srcNum))
          end if
        end do
      end if

      where(mask)&
        srcData%bkgFluxPerPixel = averageBkgFluxPerPixel

    end if

    !...................................................................
    ! Correct bkg flux for pixel size:

!    srcPixelAreaArcsec = abs(srcWcs%X%pixelSizeWorld* srcWcs%Y%pixelSizeWorld)&
!      * 3600.0 * 3600.0
srcPixelAreaArcsec = wcsPixelAreaWorld(srcWcs) * 3600.0 * 3600.0 ! in ssclib/wcs_aux

!**** note this depends on the worldUnits being both 'deg'. Should check this.

    bkgFluxesPerSquareArcsec = srcData%bkgFluxPerPixel / srcPixelAreaArcsec

    !...................................................................
    ! Get src fov coords:

!    call raDecToInst(srcData%raDeg, srcData%decDeg, scAttitude&
!      , expStartTimeStamp, srcData%xDet, srcData%yDet, srcData%thetaArcsec&
!      , srcData%phi)
call raDecToInst(srcData%raDeg*PiOn180, srcData%decDeg*PiOn180, scAttitude&
  , expStartTimeStamp, srcData%xDet, srcData%yDet, srcData%thetaArcsec&
  , srcData%phi)
    ! in ssclib/coordinate.

    !...................................................................
    ! Get peak fluxes:

    do srcNum = 1, numSources
      integral = getPsfIntegral(srcData(srcNum)%thetaArcsec&
        , srcData(srcNum)%EeV)

      peakFluxPerSquareArcsec(srcNum) = srcData(srcNum)%flux / integral
    end do

    !...................................................................
    ! Get ellipse axes:

    call getPsfEllipseAxes(srcData%thetaArcsec, srcData%EeV&
      , peakFluxPerSquareArcsec&
      , bkgFluxesPerSquareArcsec, fracBackground, ellipseLongAxisArcsec&
      , ellipseShrtAxisArcsec, srcData%isValid)

!**** check and report about invalid sources?

    !...................................................................
    ! Convert ra and dec to image tangent coords:

!    call raDecToTan(imgWcs%X%refWorld, imgWcs%Y%refWorld, srcData%raDeg&
!      , srcData%decDeg, srcData%xTan, srcData%yTan) ! in ssclib/coordinate.
call raDecToTan(srcData%raDeg*PiOn180, srcData%decDeg*PiOn180&
  , imgWcs%X%refWorld*PiOn180, imgWcs%Y%refWorld*PiOn180&
  , srcData%xTan, srcData%yTan) ! in ssclib/coordinate.

!    call instToTan(ctrFovXtan,ctrFovYtan, imgWcs%X%refWorld, imgWcs%Y%refWorld&
!      , scAttitude, expStartTimeStamp, thetaArcsec=0.0d0, phi=0.0d0)
call telToTan(0.0d0, 0.0d0, imgWcs%X%refWorld*PiOn180&
  , imgWcs%Y%refWorld*PiOn180, scAttitude, expStartTimeStamp&
  , ctrFovXtan, ctrFovYtan)
    ! in ssclib/coordinate.

    ellipseShrtAxisTan = ellipseShrtAxisArcsec / 3600.0 * PiOn180
    ellipseLongAxisTan = ellipseLongAxisArcsec / 3600.0 * PiOn180
    ellipseRotAngle = atan2(srcData%yTan - ctrFovYtan&
      , srcData%xTan - ctrFovXtan)

    !...................................................................
    ! Convert ellipses from axes and rotAngleDeg to amps and phase:

    do srcNum = 1, numSources
      call ellipseAngleToPhase(ellipseShrtAxisTan(srcNum)&
        , ellipseLongAxisTan(srcNum), ellipseRotAngle(srcNum)&
        , ellipseXampTan(srcNum), ellipseYampTan(srcNum)&
        , srcData(srcNum)%region%phase) ! in ssclib/ellipse
    end do

    !...................................................................
    ! Change image tangents to pixel coords, also scale ellipse amps to
    ! pixel coords:

    call tanToPixels(imgWcs, srcData%xTan, srcData%yTan, srcData%xPixel&
      , srcData%yPixel) ! in ssclib/wcs_aux
!    srcData%region%xAmp = ellipseXampTan / PiOn180&
!      / abs(imgWcs%X%pixelSizeWorld)
!    srcData%region%yAmp = ellipseYampTan / PiOn180&
!      / abs(imgWcs%Y%pixelSizeWorld)
!!*** probably _shouldn't_ have the abs()s here. They also should be got rid of in the tan<->pixels converters in wcs_aux.

pixelSizeXWorld = sqrt(imgWcs%x%matrixOnAxis * imgWcs%x%matrixOnAxis&
 + imgWcs%x%matrixOffAxis * imgWcs%x%matrixOffAxis)
pixelSizeYWorld = sqrt(imgWcs%y%matrixOnAxis * imgWcs%y%matrixOnAxis&
 + imgWcs%y%matrixOffAxis * imgWcs%y%matrixOffAxis)
!*** this is only valid if there is no skew.

srcData%region%xAmp = ellipseXampTan&
  / pixelSizeXWorld / PiOn180
srcData%region%yAmp = ellipseYampTan&
  / pixelSizeYWorld / PiOn180
!**** There are several things wrong with this. Easiest to fix or at least trap is that it assumes worldUnits are both 'deg' and that ctype = 'RA---TAN/DEC--TAN' etc. Harder to trap is that the above assumes there is no skew between pixel and world coords. Finally (altho it's a related issue, if I found a way to handle skew it would also handle this) the parity of the transform gets lost in the way that pixelSizeXWorld are calculated. pixelSizeXWorld and pixelSizeYWorld end up as always positive.

    srcData%region%unit = 'pixels'

    !...................................................................
    ! Convert ellipses from amps and phase to axes and rotAngleDeg:

    do srcNum = 1, numSources
      call ellipsePhaseToAngle(&
        srcData(srcNum)%region%xAmp,&
        srcData(srcNum)%region%yAmp,&
        srcData(srcNum)%region%phase,&
        srcData(srcNum)%region%shortSemiAxis,&
        srcData(srcNum)%region%longSemiAxis,&
        srcData(srcNum)%region%rotatedAngle&
      ) ! in ssclib/ellipse
    end do

  end subroutine getSourceCutoutsFewArgs

  !---------------------------------------------------------------------
  subroutine getSourceCutoutsManyArgs(imgWcs, srcWcs, scAttitude&
  , expStartTimeStamp, fracBackground, srcEeV, srcRaDeg, srcDecDeg, srcRate&
  , srcBkgRatePixels, xPixels, yPixels, ellipseXampPixels, ellipseYampPixels&
  , ellipsePhase, srcIsValid)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up with the correct instrument beforehand.
  !
  ! NOTE: See description in getSourceCutoutsFewArgs() for detail.
  !  This routine accepts many arrays, instead of SrcDataT, for the argument,
  !  and that is the only difference from getSourceCutoutsFewArgs().
  !  In my humble opinion, to use getSourceCutoutsFewArgs() is
  !  highly recommended, rather than using this clumsy
  !  getSourceCutoutsManyArgs().

    implicit none

    type(WcsT),                   intent(in) :: imgWcs,&
                                                srcWcs
    real(double),                 intent(in) :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    real(single),                 intent(in) :: fracBackground
    real(double),                 intent(in) :: srcEeV(:),&
                                                srcRaDeg(size(srcEeV)),&
                                                srcDecDeg(size(srcEeV))
    real(single),                 intent(in) :: srcRate(size(srcEeV)),&
                                                srcBkgRatePixels(size(srcEeV))
    real(single),  intent(out) :: xPixels(size(srcEeV)),&
                                  yPixels(size(srcEeV)),&
                                  ellipseXampPixels(size(srcEeV)),&
                                  ellipseYampPixels(size(srcEeV)),&
                                  ellipsePhase(size(srcEeV))
!    logical(bool), intent(out) :: srcIsValid(size(srcEeV))
    logical, intent(out) :: srcIsValid(size(srcEeV))

    type(SrcDataT), allocatable :: srcData(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(srcData(size(srcEeV)))

    srcData%EeV             = srcEeV
    srcData%raDeg           = srcRaDeg
    srcData%decDeg          = srcDecDeg
    srcData%flux            = srcRate
    srcData%bkgFluxPerPixel = srcBkgRatePixels

    call getSourceCutoutsFewArgs(imgWcs, srcWcs, scAttitude&
      , expStartTimeStamp, fracBackground, srcData)

    xPixels           = srcData%xPixel
    yPixels           = srcData%yPixel
    ellipseXampPixels = srcData%region%xAmp
    ellipseYampPixels = srcData%region%yAmp
    ellipsePhase      = srcData%region%phase
    srcIsValid        = srcData%isValid

  end subroutine getSourceCutoutsManyArgs


  !---------------------------------------------------------------------
  subroutine dumpSrcDataTScalar(srcData, dumpVarMsg)
  !---------------------------------------------------------------------

    implicit none

    type(SrcDataT), intent(in) :: srcData
    character(*), intent(in), optional :: dumpVarMsg

    !-------------------------------------------------------------------
    ! Start

    if (present(dumpVarMsg)) then
      write(*,'("== Dump SrcDataT: ",A)') trim(dumpVarMsg)
    else
      write(*,'("== Dump SrcDataT: ")')
    end if
    write(*,'("idNum : ",I6)')   srcData%idNum	! integer(int16)
    write(*,'("flux  : ",ES14.7)') srcData%flux	! real(single)
    write(*,'("bkgFluxPerPixel : ",ES14.7)') srcData%bkgFluxPerPixel	! real(single)
    write(*,'("(xPixel,yPixel) : (",F12.6,", ",F12.6,")")') srcData%xPixel, srcData%yPixel	! real(single)
    write(*,'("fluxUnit : ",A10)') trim(srcData%fluxUnit)	! character(10)

    if (present(dumpVarMsg)) then
      call dumpEllipseT(srcData%region, trim(dumpVarMsg)//"%region")	! defined in ellipse
    else
      call dumpEllipseT(srcData%region, "SrcDataT%region")		! defined in ellipse
    end if

    write(*,'("(xDet,yDet) : (",ES21.14,", ",ES21.14,")")') srcData%xDet, srcData%yDet	! real(double)
    write(*,'("(raDeg,decDeg) : (",ES21.14,", ",ES21.14,")")') srcData%raDeg, srcData%decDeg	! real(double)
    write(*,'("EeV : ",ES14.7)') srcData%EeV	! real(double)
    write(*,'("(thetaArcsec,phi) : (",ES21.14,", ",ES21.14,")")') &
       srcData%thetaArcsec, srcData%phi	! real(double)
    write(*,'("(xTan,yTan) : (",ES21.14,", ",ES21.14,")")') srcData%xTan,srcData%yTan	! real(double)
    write(*,'("isValid : ",L1)') srcData%isValid	! logical(bool)

    write(*,'("=================")')

  end subroutine dumpSrcDataTScalar


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module source_cutouts
